 #   Copyright 2019 AUI, Inc. Washington DC, USA
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


import numpy as np
import xarray as xr
import copy
from ._sirius_utils._beam_utils import _calc_ant_jones, _calc_resolution
from ._sirius_utils._calc_parallactic_angles import _calc_parallactic_angles, _find_optimal_set_angle
from ._parm_utils._check_beam_parms import _check_beam_parms

def calc_zpc_beam(zpc_xds,parallactic_angles,freq_chan,beam_parms,check_parms=True):
    """
    Calculates an antenna apertures from Zernike polynomial coefficients, and then Fourier transform it to obtain the antenna beam image.
    The beam image dimensionality is [pa (paralactic angle), chan (channel), pol (polarization), l (orthographic/synthesis projection of directional cosine), m (orthographic/synthesis projection of directional cosine)].

    Parameters
    ----------
    zpc_xds: xr.Dataset
        A Zernike polynomial coefficient xr.Datasets. Available models can be found in sirius_data/zernike_dish_models/data.
    parallactic_angles: float np.array, radians
        An array of the parallactic angles for which to calculate the antenna beams.
    freq_chan: float np.array, [n_chan], Hz
        Channel frequencies.
    beam_parms: dict
    beam_parms['mueller_selection']: int np.array, default=np.array([ 0, 5, 10, 15])
        The elements in the 4x4 beam Mueller matrix to use. The elements are numbered row wise.
        For example [ 0, 5, 10, 15] are the diagonal elements.
    beam_parms['pa_radius']: float, default=0.2, radians
        The change in parallactic angle that will trigger the calculation of a new beam when using Zernike polynomial aperture models.
    beam_parms['image_size']: int np.array, default=np.array([1000,1000])
        Size of the beam image generated from the Zernike polynomial coefficients.
    beam_parms['fov_scaling']: int, default=15
        Used to scale the size of the beam image, which is given by fov_scaling*(1.22 *c/(dish_diam*frequency)).
    beam_parms['zernike_freq_interp']: str, default='nearest', options=['linear', 'nearest', 'zero', 'slinear', 'quadratic', 'cubic']
        What interpolation method to use for Zernike polynomial coefficients.
    check_parms: bool
        Check input parameters and asign defaults.
        
    Returns
    -------
    J_xds: xr.Dataset
        An xds that contains the image of the per antenna beam as a function of [pa (paralactic angle), chan (channel), pol (polarization), l (orthographic/synthesis projection of directional cosine), m (orthographic/synthesis projection of directional cosine)]. Should not be confused with the primary beam, which is the beam for a baseline and is equal to the product of two antenna beams.
    """
    _beam_parms = copy.deepcopy(beam_parms)
    
    if check_parms: assert(_check_beam_parms(_beam_parms)), "######### ERROR: beam_parms checking failed."
    
    
    pb_freq = freq_chan
    pb_pa = parallactic_angles
    
    min_delta = _calc_resolution(pb_freq,zpc_xds.dish_diam,_beam_parms)
    #print('min_delta',min_delta)
    _beam_parms['cell_size'] = np.array([-min_delta,min_delta]) #- sign?
  
    map_mueler_to_pol = np.array([[0,0],[0,1],[1,0],[1,1],[0,2],[0,3],[1,2],[1,3],[2,0],[2,1],[3,0],[3,1],[2,2],[2,3],[3,2],[3,3]])
    _beam_parms['needed_pol'] = np.unique(np.ravel(map_mueler_to_pol[_beam_parms['mueller_selection']]))
    
    assert (0 in _beam_parms['mueller_selection']) or (15 in _beam_parms['mueller_selection']), "Mueller element 0 or 15 must be selected."
    
    pb_planes = _calc_ant_jones(zpc_xds,pb_freq,pb_pa,_beam_parms)
    
    image_size = _beam_parms['image_size']
    image_center = image_size//2
    cell_size = _beam_parms['cell_size']
    
    image_center = np.array(image_size)//2
    l = np.arange(-image_center[0], image_size[0]-image_center[0])*cell_size[0]
    m = np.arange(-image_center[1], image_size[1]-image_center[1])*cell_size[1]
    
    coords = {'chan':pb_freq, 'pa': pb_pa, 'pol': _beam_parms['needed_pol'],'l':l,'m':m}

    J_xds = xr.Dataset()
    J_xds = J_xds.assign_coords(coords)
    
    J_xds['J'] = xr.DataArray(pb_planes, dims=['pa','chan','pol','l','m'])
    
    return J_xds
    

def evaluate_beam_models(beam_models,time_str,freq_chan,phase_center_ra_dec,site_location,beam_parms,check_parms=True):
    """
    Loops over beam_models and converts each Zernike polynomial coefficient xr.Datasets to an antenna beam image. The beam image dimensionality is [pa (paralactic angle), chan (channel), pol (polarization), l (orthographic/synthesis projection of directional cosine), m (orthographic/synthesis projection of directional cosine)]. The parallactic angles are also calculated for each date-time in time_str at the site_location and with a right ascension declination in phase_center_ra_dec. A subset of parallactic angles are used in the pa coordinate of the beam image, where all pa values are within beam_parms['pa_radius'] radians.

    Parameters
    ----------
    beam_models: list
        List of beam models to use. Beam models can be any combination of function parameter dictionaries, image xr.Datasets or Zernike polynomial coefficient xr.Datasets (models can be found in sirius_data/zernike_dish_models/data).
    time_str: str np.array, [n_time], 'YYYY-MM-DDTHH:MM:SS.SSS'
        Time series. Example '2019-10-03T19:00:00.000'.
    freq_chan: float np.array, [n_chan], Hz
        Channel frequencies.
    phase_center_ra_dec: float np.array, [n_time, 2], (singleton: n_time), radians
        Phase center of array.
    site_location: dict
        A dictionary with the location of telescope. For example [{'m0': {'unit': 'm', 'value': -1601185}, 'm1': {'unit': 'm', 'value': -5041977}, 'm2': {'unit': 'm', 'value': 3554875}, 'refer': 'ITRF', 'type': 'position'}]. The site location of telescopes can be found in site_pos attribute of the xarray dataset of the radio telescope array layout (see zarr files in sirius_data/telescope_layout/data/).
    parallactic_angles: float np.array, radians
        An array of the parallactic angles for which to calculate the antenna beams.
    freq_chan: float np.array, [n_chan], Hz
        Channel frequencies.
    beam_parms: dict
    beam_parms['mueller_selection']: int np.array, default=np.array([ 0, 5, 10, 15])
        The elements in the 4x4 beam Mueller matrix to use. The elements are numbered row wise.
        For example [ 0, 5, 10, 15] are the diagonal elements.
    beam_parms['pa_radius']: float, default=0.2, radians
        The change in parallactic angle that will trigger the calculation of a new beam when using Zernike polynomial aperture models.
    beam_parms['image_size']: int np.array, default=np.array([1000,1000])
        Size of the beam image generated from the Zernike polynomial coefficients.
    beam_parms['fov_scaling']: int, default=15
        Used to scale the size of the beam image, which is given by fov_scaling*(1.22 *c/(dish_diam*frequency)).
    beam_parms['zernike_freq_interp']: str, default='nearest', options=['linear', 'nearest', 'zero', 'slinear', 'quadratic', 'cubic']
        What interpolation method to use for Zernike polynomial coefficients.
        
    Returns
    -------
    J_xds: xr.Dataset
        An xds that contains the image of the per antenna beam as a function of [pa (paralactic angle), chan (channel), pol (polarization), l (orthographic/synthesis projection of directional cosine), m (orthographic/synthesis projection of directional cosine)]. Should not be confused with the primary beam, which is the beam for a baseline and is equal to the product of two antenna beams.
    """
    
    #Calculate parallactic angles.
    pa = _calc_parallactic_angles(time_str,site_location,phase_center_ra_dec)
    pa_subset,vals_dif = _find_optimal_set_angle(pa[:,None],beam_parms['pa_radius'] )
    
    _beam_parms = copy.deepcopy(beam_parms)

    # If beam model is a Zernike polynomial coefficient xr.Datasets convert it to an image.
    eval_beam_models = []
    for bm in beam_models:
        if 'ZC' in bm: #check for zpc files
            J_xds = calc_zpc_beam(bm,pa_subset,freq_chan,_beam_parms,check_parms)
            J_xds.attrs = bm.attrs
            eval_beam_models.append(J_xds)
        else:
            eval_beam_models.append(bm)
    
    return eval_beam_models, pa



#   Copyright 2019 AUI, Inc. Washington DC, USA
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from ._parm_utils._check_uvw_parms import _check_uvw_parms
from ._sirius_utils._array_utils import _calc_baseline_indx_pair
from ._sirius_utils._uvw_utils import _calc_uvw_astropy, _calc_uvw_casacore, _calc_uvw_casa
import copy
import numpy as np


def calc_uvw_chunk(tel_xds, time_str, phase_center_ra_dec, uvw_parms, check_parms=True):
    """
    Calculates a chunk of uvw values. This function forms part of a node in a Dask graph. 
    This function can be used on its own if no parallelism is required.
    
    Parameters
    ----------
    tel_xds: xr.Dataset
        An xarray dataset of the radio telescope array layout (see zarr files in sirius_data/telescope_layout/data/ for examples). 
    time_str: str np.array, [n_time], 'YYYY-MM-DDTHH:MM:SS.SSS'
        Time serie. Example '2019-10-03T19:00:00.000'.
    phase_center_ra_dec: numpy.array of floats, [n_time, 2], (singleton: n_time), radians
        Phase center of array.
    uvw_parms: dict
    uvw_parms['calc_method']: str, default='astropy', options=['astropy','casa']
        The uvw coordinates can be calculated using the astropy package or the casa measures tool.
    uvw_parms['auto_corr']: bool, default=False
        If True autocorrelations are also calculated.
    check_parms: bool
        Check input parameters and asign defaults.
    
    Returns
    -------
    uvw: float np.array, [n_time,n_baseline,3]
        The uvw coorsdinates in per wavelength units.
    antenna1: int np.array, [n_baseline]
        The indices of the first antenna in the baseline pairs.
    antenna2: int np.array, [n_baseline]
        The indices of the second antenna in the baseline pairs.
    """
    
    _uvw_parms = copy.deepcopy(uvw_parms)
    if check_parms: assert(_check_uvw_parms(_uvw_parms)), "######### ERROR: calc_uvw uvw_parms checking failed."
        
    n_ant = tel_xds.dims['ant_name']
    antenna1,antenna2=_calc_baseline_indx_pair(n_ant,_uvw_parms['auto_corr']) #The indices of the antenna pairs for each baseline.
                                                   
    if _uvw_parms['calc_method'] == 'astropy':
        uvw = _calc_uvw_astropy(tel_xds, time_str, phase_center_ra_dec, antenna1, antenna2)
    elif _uvw_parms['calc_method'] == 'casa': #CASA measures tool is not thread safe so python casacore is the prefered CASA method.
        uvw = _calc_uvw_casacore(tel_xds, time_str, phase_center_ra_dec, antenna1, antenna2)
    elif _uvw_parms['calc_method'] == 'casa_thread_unsafe': #This function is retained for future development. 
        uvw = _calc_uvw_casa(tel_xds, time_str, phase_center_ra_dec, antenna1, antenna2)
    #elif _uvw_parms['calc_method'] == 'calc11': #Under development.
        #uvw = _calc_uvw_casa(tel_xds, time_str, phase_center_ra_dec, antenna1, antenna2)
    return uvw, antenna1, antenna2

''' Under development 
def calc_uvw(tel_xds, time_str, phase_center_ra_dec, uvw_parms, check_parms=True):
    """
    Under development.
    
    Parameters
    ----------
    
    Returns
    -------
    """
    _uvw_parms = copy.deepcopy(uvw_parms)
    _save_parms = copy.deepcopy(save_parms)
    assert(_check_uvw_parms(_uvw_parms)), "######### ERROR: calc_uvw uvw_parms checking failed."
 
    n_time = len(time_xda)
    n_ant = tel_xds.dims['ant_name']
    #print(n_time,n_chan,n_ant)
   
    n_time_chunks = time_xda.data.numblocks[0]
                
    uvw_list = _ndim_list((n_time_chunks,1,1))
                
    from ._sirius_utils._array_utils import _calc_n_baseline
    n_baselines = _calc_n_baseline(n_ant,_uvw_parms['auto_corr'])
                
    # Build graph
    for c_time in range(n_time_chunks):
                    #print(c_time,c_chan)
        time_chunk = time_xda.data.partitions[c_time]
        
        sim_chunk = dask.delayed(calc_uvw_chunk)(
            dask.delayed(point_source_flux_chunk),
            dask.delayed(point_source_ra_dec_chunk),
            dask.delayed(pointing_ra_dec_chunk),
            dask.delayed(phase_center_ra_dec_chunk),
            dask.delayed(_beam_parms),beam_models,
            dask.delayed(beam_model_map),
            dask.delayed(_uvw_parms),
            tel_xds,
            time_chunk,
            chan_chunk,
            dask.delayed(pol), dask.delayed(_noise_parms),
            dask.delayed(None))
        uvw_list[c_time][0][0] = da.from_delayed(sim_chunk[1],(len(time_chunk), n_baselines, 3),dtype=np.complex)

        

    uvw = da.block(uvw_list)

        
    if _save_parms['DAG_name_uvw_gen']:
        dask.visualize([vis,uvw],filename=_save_parms['DAG_name_uvw_gen'])
    
    #Create simple xds with simulated vis, uvw, weight and sigma
    uvw_xds = xr.Dataset()
    coords = {'time':time_xda.data,'chan': chan_xda.data, 'pol': pol}
    uvw_xds = vis_xds.assign_coords(coords)
        

    uvw_xds['UVW'] = xr.DataArray(uvw, dims=['time','baseline','uvw'])
    ###################
        
        write_to_ms(vis_xds, time_xda, chan_xda, pol, tel_xds, phase_center_names, phase_center_ra_dec, _uvw_parms['auto_corr'],_save_parms)
        
        return uvw_xds
'''



 #   Copyright 2019 AUI, Inc. Washington DC, USA
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import os
import time
import numpy as np
import dask.array as da
import dask
import xarray as xr
from itertools import cycle
import itertools
import copy
from astropy import units as u
from ._sirius_utils._array_utils import _ndim_list, _calc_n_baseline, _is_subset
from ._parm_utils._check_beam_parms import _check_beam_parms
from ._parm_utils._check_uvw_parms import _check_uvw_parms
from ._parm_utils._check_save_parms import _check_save_parms
from ._parm_utils._check_noise_parms import _check_noise_parms
from sirius_data._constants import pol_codes_RL, pol_codes_XY
from sirius.calc_a_noise import calc_a_noise_chunk
from sirius.calc_uvw import calc_uvw_chunk 
from sirius.calc_vis import calc_vis_chunk
from sirius.calc_beam import evaluate_beam_models
from sirius.dio import write_to_ms


def simulation(point_source_flux, point_source_ra_dec, pointing_ra_dec, phase_center_ra_dec, phase_center_names, beam_parms,beam_models,beam_model_map,uvw_parms, tel_xds, time_xda, chan_xda, pol, noise_parms, save_parms):
    """
    Creates a dask graph that computes a simulated measurement set and triggers a compute and saves the ms to disk.
    
    Parameters
    ----------
    point_source_flux: float np.array, [n_point_sources,n_time, n_chan, n_pol], (singleton: n_time, n_chan), Janskys
        The flux of the point sources.
    point_source_ra_dec: float np.array, [n_time, n_point_sources, 2], (singleton: n_time), radians
        The position of the point sources.
    pointing_ra_dec: float np.array, [n_time, n_ant, 2], (singleton: n_time, n_ant), radians
        Pointings of antennas, if they are different from the phase center. Set to None if no pointing offsets are required.
    phase_center_ra_dec: float np.array, [n_time, 2], (singleton: n_time), radians
        Phase center of array.
    phase_center_names: str np.array, [n_time], (singleton: n_time)
        Strings that are used to label phase centers.
    beam_parms: dict
    beam_parms['mueller_selection']: int np.array, default=np.array([ 0, 5, 10, 15])
        The elements in the 4x4 beam Mueller matrix to use. The elements are numbered row wise.
        For example [ 0, 5, 10, 15] are the diagonal elements.
    beam_parms['pa_radius']: float, default=0.2, radians
        The change in parallactic angle that will trigger the calculation of a new beam when using Zernike polynomial aperture models.
    beam_parms['image_size']: int np.array, default=np.array([1000,1000])
        Size of the beam image generated from the Zernike polynomial coefficients.
    beam_parms['fov_scaling']: int, default=15
        Used to scale the size of the beam image which is given fov_scaling*(1.22 *c/(dish_diam*frequency)).
    beam_parms['zernike_freq_interp']: str, default='nearest', options=['linear', 'nearest', 'zero', 'slinear', 'quadratic', 'cubic']
        What interpolation method to use for Zernike polynomial coefficients.
    beam_models: list
        List of beam models to use. Beam models can be any combination of function parameter dictionaries, image xr.Datasets or Zernike polynomial coefficient xr.Datasets.
    beam_model_map: int np.array, [n_ant]
        Each element in beam_model_map is an index into beam_models.
    uvw_parms: dict
    uvw_parms['calc_method']: str, default='astropy', options=['astropy','casa']
        Astropy coordinates or CASA tool measures can be used to calculate uvw coordinates.
    uvw_parms['auto_corr']: bool, default=False
        If True autocorrelations are also calculated.
    tel_xds: xr.Dataset
        An xarray dataset of the radio telescope array layout (see zarr files in sirius_data/telescope_layout/data/ for examples). 
    time_xda: xr.DataArray
        Time series xarray array.
    chan_xda: xr.DataArray
        Channel frequencies xarray array.
    pol: int np.array 
        Must be a subset of ['RR','RL','LR','LL'] => [5,6,7,8] or ['XX','XY','YX','YY'] => [9,10,11,12].
    noise_parms: dict
        Set various system parameters from which the thermal (ie, random additive) noise level will be calculated.
        See https://casadocs.readthedocs.io/en/stable/api/tt/casatools.simulator.html#casatools.simulator.simulator.setnoise.
    noise_parms['mode']: str, default='tsys-manuel', options=['simplenoise','tsys-manuel','tsys-atm']
        Currently only 'tsys-manuel' is implemented.
    noise_parms['t_atmos']: , float, default = 250.0, Kelvin
        Temperature of atmosphere (mode='tsys-manual')
    noise_parms['tau']: float, default = 0.1
        Zenith Atmospheric Opacity (if tsys-manual). Currently the effect of Zenith Atmospheric Opacity (Tau) is not included in the noise modeling.
    noise_parms['ant_efficiency']: float, default=0.8
        Antenna efficiency
    noise_parms['spill_efficiency']: float, default=0.85
        Forward spillover efficiency.
    noise_parms['corr_efficiency']: float, default=0.88
        Correlation efficiency.
    noise_parms['t_receiver']: float, default=50.0, Kelvin
        Receiver temp (ie, all non-atmospheric Tsys contributions).
    noise_parms['t_ground']: float, default=270.0, Kelvin
        Temperature of ground/spill.
    noise_parms['t_cmb']: float, default=2.725, Kelvin
        Cosmic microwave background temperature.
    save_parms: dict
    save_parms['write_to_ms']: bool, default=True
        Name of the ms file to create. If None no compute is triggered and an xds that contains the graph (lazy) that calculates the visibilities, uvw coordinates, weights and timing information is returned.
    save_parms['DAG_name_vis_uvw_gen']: str, default=False
        Creates a DAG diagram png, named save_parms['DAG_name_write'], of how the visibilities and uvw coordinates are calculated.
    save_parms['DAG_name_write']: str, default=False
        Creates a DAG diagram png, named save_parms['DAG_name_write'], of how the ms is created with name.
    save_parms['ms_name']:str, default='sirius_sim.ms'
    
    Returns
    -------
    ms_xds: xr.Dataset
        If save_parms['ms_name'] is not None a dask ms xds is retruned (a compute is triggered and the ms is saved to disk). 
        If save_parms['ms_name'] is None an xds that contains the graph (lazy) that calculates the visibilities, uvw coordinates, weights and timing information is returned.
    """
    
    ########################
    ### Check Parameters ###
    ########################
    
    _beam_parms = copy.deepcopy(beam_parms)
    _uvw_parms = copy.deepcopy(uvw_parms)
    _save_parms = copy.deepcopy(save_parms)
    _noise_parms = copy.deepcopy(noise_parms)
    assert(_check_uvw_parms(_uvw_parms)), "######### ERROR: calc_uvw uvw_parms checking failed."
    assert(_check_beam_parms(_beam_parms)), "######### ERROR: beam_parms checking failed."
    if noise_parms is not None:
        _noise_parms['freq_resolution'] = chan_xda.freq_resolution
        _noise_parms['time_delta'] = time_xda.time_delta
        _noise_parms['auto_corr'] = _uvw_parms['auto_corr']
        assert(_check_noise_parms(_noise_parms)), "######### ERROR: beam_parms checking failed."
    assert(_check_save_parms(_save_parms)), "######### ERROR: save_parms checking failed."
    
    pol = np.array(pol)
    assert(_is_subset(pol_codes_RL,pol) or _is_subset(pol_codes_XY,pol)), 'Pol selection invalid, must either be subset of [5,6,7,8] or [9,10,11,12] but is '
    
    #Get dimensions of data.
    n_time = len(time_xda)
    n_ant = tel_xds.dims['ant_name']
    n_baselines = _calc_n_baseline(n_ant,_uvw_parms['auto_corr'])
    n_chan = len(chan_xda)
    n_pol = len(pol)
    
    #Check dimensions.
    assert(point_source_flux.shape[0] == point_source_ra_dec.shape[1]), 'n_point_sources dimension of point_source_flux[' + str(point_source_flux.shape[0]) +'] and point_source_ra_dec['+str(point_source_ra_dec.shape[1])+'] do not match.'
    assert(point_source_flux.shape[1] == 1) or (point_source_flux.shape[1] == n_time), 'n_time dimension in point_source_flux[' + str(point_source_flux.shape[1]) + '] must be either 1 or ' + str(n_time) + ' (see time_xda parameter).'
    assert(point_source_flux.shape[2] == 1) or (point_source_flux.shape[2] == n_chan), 'n_chan dimension in point_source_flux[' + str(point_source_flux.shape[2]) + '] must be either 1 or ' + str(n_chan) + ' (see chan_xda parameter).'
    assert(point_source_flux.shape[3] == 4), 'n_pol dimension in point_source_flux[' + str(point_source_flux.shape[3]) + '] must be 4.'
    
    assert(point_source_ra_dec.shape[0] == 1) or (point_source_ra_dec.shape[0] == n_time), 'n_time dimension in point_source_ra_dec[' + str(point_source_ra_dec.shape[0]) + '] must be either 1 or ' + str(n_time) + ' (see time_xda parameter).'
    assert(point_source_ra_dec.shape[2] == 2), 'ra,dec dimension in point_source_ra_dec[' + str(point_source_ra_dec.shape[2]) + '] must be 2.' 
    
    if pointing_ra_dec is not None:
        assert(pointing_ra_dec.shape[0] == 1) or (pointing_ra_dec.shape[0] == n_time), 'n_time dimension in pointing_ra_dec[' + str(pointing_ra_dec.shape[0]) + '] must be either 1 or ' + str(n_time) + ' (see time_xda parameter).'
        assert(pointing_ra_dec.shape[1] == 1) or (pointing_ra_dec.shape[1] == n_ant), 'n_ant dimension in pointing_ra_dec[' + str(pointing_ra_dec.shape[1]) + '] must be either 1 or ' + str(n_ant) + ' (see tel_xds.dims[\'ant_name\']).'
        assert(pointing_ra_dec.shape[2] == 2), 'ra,dec dimension in pointing_ra_dec[' + str(pointing_ra_dec.shape[2]) + '] must be 2.'
        
        
    assert(phase_center_ra_dec.shape[0] == 1) or (phase_center_ra_dec.shape[0] == n_time), 'n_time dimension in phase_center_ra_dec[' + str(phase_center_ra_dec.shape[0]) + '] must be either 1 or ' + str(n_time) + ' (see time_xda parameter).'
    assert(phase_center_ra_dec.shape[1] == 2), 'ra,dec dimension in phase_center_ra_dec[' + str(phase_center_ra_dec.shape[1]) + '] must be 2.' 
        
    assert(phase_center_names.shape[0] == 1) or (phase_center_names.shape[0] == n_time), 'n_time dimension in phase_center_ra_dec[' + str(phase_center_names.shape[0]) + '] must be either 1 or ' + str(n_time) + ' (see time_xda parameter).'
    
    #Find all singleton dimensions, so that indexing is done correctly.
    f_pc_time = n_time if phase_center_ra_dec.shape[0] == 1 else 1
    f_ps_time = n_time if point_source_ra_dec.shape[0] == 1 else 1
    f_sf_time = n_time if point_source_flux.shape[1] == 1 else 1
    f_sf_chan = n_chan if point_source_flux.shape[2] == 1 else 1
    
    do_pointing = False
    if pointing_ra_dec is not None:
        do_pointing = True
        f_pt_time = n_time if phase_center_ra_dec.shape[0] == 1 else 1
        f_pt_ant =  n_ant if point_source_ra_dec.shape[1] == 1 else 1
    else:
        pointing_ra_dec = np.zeros((2,2,2))
        f_pt_time = n_time
        f_pt_ant = n_ant
    
    ###################
    ### Build graph ###
    ###################
    
    # Number of parallel processes will be equal to n_time_chunks x n_chan_chunks.
    n_time_chunks = time_xda.data.numblocks[0]
    n_chan_chunks = chan_xda.data.numblocks[0]
    
    #Iter over time,chan
    iter_chunks_indx = itertools.product(np.arange(n_time_chunks), np.arange(n_chan_chunks))
    n_pol = len(pol)
    
    #Create empty n-dimensional lists where delayed arrays will be stored.
    vis_list = _ndim_list((n_time_chunks,1,n_chan_chunks,1))
    uvw_list = _ndim_list((n_time_chunks,1,1))
    weight_list = _ndim_list((n_time_chunks,1,1))
    sigma_list = _ndim_list((n_time_chunks,1,1))
    t_list = _ndim_list((n_time_chunks,n_chan_chunks,1))
    
    for c_time, c_chan in iter_chunks_indx:
        #print(c_time,c_chan)
        time_chunk = time_xda.data.partitions[c_time]
        chan_chunk = chan_xda.data.partitions[c_chan]
        
        s_time = c_time*time_xda.data.chunks[0][0]
        e_time = c_time*time_xda.data.chunks[0][0] + time_xda.data.chunks[0][c_time] - 1 #-1 needed for // to work.
        s_chan = c_chan*chan_xda.data.chunks[0][0]
        e_chan = c_chan*chan_xda.data.chunks[0][0] + chan_xda.data.chunks[0][c_chan] - 1 #-1 needed for // to work.
        
        point_source_flux_chunk = point_source_flux[:,s_time//f_sf_time:e_time//f_sf_time+1,s_chan//f_sf_chan:e_chan//f_sf_chan+1,:]
        point_source_ra_dec_chunk = point_source_ra_dec[s_time//f_ps_time:e_time//f_ps_time+1,:,:]
        phase_center_ra_dec_chunk = phase_center_ra_dec[s_time//f_pc_time:e_time//f_pc_time+1,:]
        
        if do_pointing:
            pointing_ra_dec_chunk = pointing_ra_dec[s_time//f_pt_time:e_time//f_pt_time+1,:,:]
        else:
            pointing_ra_dec_chunk = None

        ### TO DO ###
        # Subselect channels for each beam_model with channel axis
        
        sim_chunk = dask.delayed(simulation_chunk)(
            dask.delayed(point_source_flux_chunk),
            dask.delayed(point_source_ra_dec_chunk),
            dask.delayed(pointing_ra_dec_chunk),
            dask.delayed(phase_center_ra_dec_chunk),
            dask.delayed(_beam_parms),beam_models,
            dask.delayed(beam_model_map),
            dask.delayed(_uvw_parms),
            tel_xds,
            time_chunk,
            chan_chunk,
            dask.delayed(pol), dask.delayed(_noise_parms),
            dask.delayed(None))

        vis_list[c_time][0][c_chan][0] = da.from_delayed(sim_chunk[0],(len(time_chunk), n_baselines, len(chan_chunk),n_pol),dtype=np.complex)
        uvw_list[c_time][0][0] = da.from_delayed(sim_chunk[1],(len(time_chunk), n_baselines, 3),dtype=np.complex)
        weight_list[c_time][0][0] = da.from_delayed(sim_chunk[2],(len(time_chunk), n_baselines, n_pol),dtype=np.float)
        sigma_list[c_time][0][0] = da.from_delayed(sim_chunk[3],(len(time_chunk), n_baselines, n_pol),dtype=np.float)
        t_list[c_time][c_chan][0] = da.from_delayed(sim_chunk[4],((4,)),dtype=np.float)
        
    vis = da.block(vis_list)
    uvw = da.block(uvw_list)
    weight = da.block(weight_list)
    sigma = da.block(sigma_list)
    timing = da.block(t_list)
    
    if _save_parms['DAG_name_vis_uvw_gen']:
        dask.visualize([vis,uvw],filename=_save_parms['DAG_name_vis_uvw_gen'])
        
    #Create simple xds with simulated vis, uvw, weight and sigma
    vis_xds = xr.Dataset()
    coords = {'time':time_xda.data,'chan': chan_xda.data, 'pol': pol}
    vis_xds = vis_xds.assign_coords(coords)
        
    vis_xds['DATA'] = xr.DataArray(vis, dims=['time','baseline','chan','pol'])
    vis_xds['UVW'] = xr.DataArray(uvw, dims=['time','baseline','uvw'])
    vis_xds['WEIGHT'] = xr.DataArray(weight, dims=['time','baseline','pol'])
    vis_xds['SIGMA'] = xr.DataArray(sigma, dims=['time','baseline','pol'])
    vis_xds['TIMING'] = xr.DataArray(timing, dims=['time_chunk','chan_chunk','4'])
        
    if save_parms['write_to_ms']:
        ms_xds= write_to_ms(vis_xds, time_xda, chan_xda, pol, tel_xds, phase_center_names, phase_center_ra_dec, _uvw_parms['auto_corr'],_save_parms)
        return ms_xds
    else:
        return vis_xds
    

def simulation_chunk(point_source_flux, point_source_ra_dec, pointing_ra_dec, phase_center_ra_dec, beam_parms,beam_models,beam_model_map,uvw_parms, tel_xds, time_chunk, chan_chunk, pol, noise_parms, uvw_precompute=None):
    """
    Simulates uvw coordinates, interferometric visibilities and adds noise. This function does not produce a measurement set.  
    
    Parameters
    ----------
    point_source_flux: float np.array, [n_point_sources,n_time, n_chan, n_pol], (singleton: n_time, n_chan), Janskys
        The flux of the point sources.
    point_source_ra_dec: float np.array, [n_time, n_point_sources, 2], (singleton: n_time), radians
        The position of the point sources.
    pointing_ra_dec: float np.array, [n_time, n_ant, 2], (singleton: n_time, n_ant), radians
        Pointings of antennas, if they are different from the phase center. Set to None if no pointing offsets are required.
    phase_center_ra_dec: float np.array, [n_time, 2], (singleton: n_time), radians
        Phase center of array.
    beam_parms: dict
    beam_parms['mueller_selection']: int np.array, default=np.array([ 0, 5, 10, 15])
        The elements in the 4x4 beam Mueller matrix to use. The elements are numbered row wise.
        For example [ 0, 5, 10, 15] are the diagonal elements.
    beam_parms['pa_radius']: float, default=0.2, radians
        The change in parallactic angle that will trigger the calculation of a new beam when using Zernike polynomial aperture models.
    beam_parms['image_size']: int np.array, default=np.array([1000,1000])
        Size of the beam image generated from the Zernike polynomial coefficients.
    beam_parms['fov_scaling']: int, default=15
        Used to scale the size of the beam image, which is given by fov_scaling*(1.22 *c/(dish_diam*frequency)).
    beam_parms['zernike_freq_interp']: str, default='nearest', options=['linear', 'nearest', 'zero', 'slinear', 'quadratic', 'cubic']
        What interpolation method to use for Zernike polynomial coefficients.
    beam_models: list
        List of beam models to use. Beam models can be any combination of function parameter dictionaries, image xr.Datasets or Zernike polynomial coefficient xr.Datasets.
    beam_model_map: int np.array, [n_ant]
        Each element in beam_model_map is an index into beam_models.
    uvw_parms: dict
    uvw_parms['calc_method']: str, default='astropy', options=['astropy','casa']
        Astropy coordinates or CASA tool measures can be used to calculate uvw coordinates.
    uvw_parms['auto_corr']: bool, default=False
        If True autocorrelations are also calculated.
    tel_xds: xr.Dataset
        An xarray dataset of the radio telescope array layout (see zarr files in sirius_data/telescope_layout/data/ for examples). 
    time_chunk: str np.array, [n_time], 'YYYY-MM-DDTHH:MM:SS.SSS'
        Time series. Example '2019-10-03T19:00:00.000'.
    chan_chunk: float np.array, [n_chan], Hz
        Channel frequencies.
    pol: int np.array 
        Must be a subset of ['RR','RL','LR','LL'] => [5,6,7,8] or ['XX','XY','YX','YY'] => [9,10,11,12].
    noise_parms: dict
        Set various system parameters from which the thermal (ie, random additive) noise level will be calculated.
        See https://casadocs.readthedocs.io/en/stable/api/tt/casatools.simulator.html#casatools.simulator.simulator.setnoise.
    noise_parms['mode']: str, default='tsys-manuel', options=['simplenoise','tsys-manuel','tsys-atm']
        Currently only 'tsys-manuel' is implemented.
    noise_parms['t_atmos']: , float, default = 250.0, Kelvin
        Temperature of atmosphere (mode='tsys-manual')
    noise_parms['tau']: float, default = 0.1
        Zenith Atmospheric Opacity (if tsys-manual). Currently the effect of Zenith Atmospheric Opacity (Tau) is not included in the noise modeling.
    noise_parms['ant_efficiency']: float, default=0.8
        Antenna efficiency
    noise_parms['spill_efficiency']: float, default=0.85
        Forward spillover efficiency.
    noise_parms['corr_efficiency']: float, default=0.88
        Correlation efficiency.
    noise_parms['t_receiver']: float, default=50.0, Kelvin
        Receiver temp (ie, all non-atmospheric Tsys contributions).
    noise_parms['t_ground']: float, default=270.0, Kelvin
        Temperature of ground/spill.
    noise_parms['t_cmb']: float, default=2.725, Kelvin
        Cosmic microwave background temperature.
    save_parms: dict
    save_parms['write_to_ms']: bool, default=True
        Name of the ms file to create. If None no compute is triggered and an xds that contains the graph (lazy) that calculates the visibilities, uvw coordinates, weights and timing information is returned.
    save_parms['DAG_name_vis_uvw_gen']: str, default=False
        Creates a DAG diagram png, named save_parms['DAG_name_write'], of how the visibilities and uvw coordinates are calculated.
    save_parms['DAG_name_write']: str, default=False
        Creates a DAG diagram png, named save_parms['DAG_name_write'], of how the ms is created with name.
    save_parms['ms_name']:str, default='sirius_sim.ms'
    
    Returns
    -------
    vis : complex np.array, [n_time,n_baseline,n_chan,n_pol]   
        Visibility data.
    uvw : float np.array, [n_time,n_baseline,3]   
        Spatial frequency coordinates.
    weight: complex np.array, [n_time,n_baseline,n_pol]
        Data weights.
    sigma: complex np.array, [n_time,n_baseline,n_pol]
        RMS noise of data.
    t_arr: float np.array, [4]
        Timing infromation: calculate uvw, evaluate_beam_models, calculate visibilities, calculate additive noise.
    """

    #Calculate uvw coordinates
    t0 = time.time()
    if uvw_precompute is None:
        uvw, antenna1,antenna2 = calc_uvw_chunk(tel_xds, time_chunk, phase_center_ra_dec, uvw_parms,check_parms=False)
    else:
        from ._sirius_utils._array_utils import _calc_baseline_indx_pair
        n_ant = len(ant_pos)
        antenna1,antenna2=_calc_baseline_indx_pair(n_ant,uvw_parms['auto_corr'])
        uvw = uvw_precompute
    t0 = time.time()-t0
      
    t1 = time.time()
    #Evaluate zpc files
    eval_beam_models, pa = evaluate_beam_models(beam_models,time_chunk,chan_chunk,phase_center_ra_dec,tel_xds.site_pos,beam_parms,check_parms=False)
    t1 = time.time()-t1

    #Calculate visibilities
    t2 = time.time()
    vis_data_shape =  np.concatenate((uvw.shape[0:2],[len(chan_chunk)],[len(pol)]))
    vis =calc_vis_chunk(uvw,vis_data_shape,point_source_flux,point_source_ra_dec,pointing_ra_dec,phase_center_ra_dec,antenna1,antenna2,chan_chunk,beam_model_map,eval_beam_models, pa, pol, beam_parms['mueller_selection'],check_parms=False)
    t2 = time.time()-t2

    #Calculate and add noise
    t3 = time.time()
    if noise_parms is not None:
        noise, weight, sigma = calc_a_noise_chunk(vis.shape,uvw,beam_model_map,eval_beam_models, antenna1, antenna2,noise_parms,check_parms=False)
        vis = vis + noise
    else:
        n_time, n_baseline, n_chan, n_pol = vis.shape
        weight = np.ones((n_time,n_baseline,n_pol))
        sigma = np.ones((n_time,n_baseline,n_pol))
    t3 = time.time()-t3
    
    t_arr = np.array([t0,t1,t2,t3])
    return vis, uvw, weight, sigma, t_arr



 #   Copyright 2019 AUI, Inc. Washington DC, USA
 #
 #   Licensed under the Apache License, Version 2.0 (the "License");
 #   you may not use this file except in compliance with the License.
 #   You may obtain a copy of the License at
 #
 #       http://www.apache.org/licenses/LICENSE-2.0
 #
 #   Unless required by applicable law or agreed to in writing, software
 #   distributed under the License is distributed on an "AS IS" BASIS,
 #   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 #   See the License for the specific language governing permissions and
 #   limitations under the License.
 
import numpy as np
import itertools
import time
import xarray as xr
import dask.array as da
import numba
from numba import jit
from sirius._sirius_utils._primary_beam_funcs import _casa_airy_pb, _airy_pb, _casa_airy_pb_njit, _airy_pb_njit
from sirius._sirius_utils._math_utils import _interp_array
from sirius._sirius_utils._coord_transforms import  _rot_coord, _compute_rot_coords
from sirius._sirius_utils._zernike_polynomials import _generate_zernike_surface
from sirius._sirius_utils._array_utils import _find_angle_indx, _find_val_indx
from sirius_data._constants import map_mueler_to_pol, c


def _calc_resolution(pb_freq,dish_diameter,beam_parms):
    fov = beam_parms['fov_scaling']*(1.22 * c / (dish_diameter*pb_freq))
    min_delta = min(min(fov/beam_parms['image_size'][0]),min(fov/beam_parms['image_size'][1]))
    return min_delta

def _beam_models_to_tuple(beam_models,beam_model_map):
    new_beam_model_map = np.zeros(beam_model_map.shape,dtype=np.int)
    beam_types = np.zeros(beam_model_map.shape,dtype=np.int)
    beam_models_list_type0 = []
    beam_models_list_type1 = []
    
    i_t0 = 0
    i_t1 = 0
    for i,bm in enumerate(beam_models):
        if "J" in bm:
            beam_models_list_type0.append(_exstract_arrays_from_bm_xds(bm))
            beam_types[beam_model_map==i] = 0
            new_beam_model_map[beam_model_map==i] = i_t0
            i_t0 = i_t0 + 1
        else:
            beam_models_list_type1.append(_exstract_vals_from_analytic_dict(bm))
            beam_types[beam_model_map==i] = 1
            new_beam_model_map[beam_model_map==i] = i_t1
            i_t1 = i_t1 + 1
            
    #Dummy for sample_j
    if not bool(beam_models_list_type0):
        d_arr_float = np.array([0.])
        d_arr_int = np.array([0])
        d_J = np.zeros((1,1,1,1,1),np.complex128)
        beam_models_list_type0.append((0,d_J,d_arr_float,d_arr_float,
                                        d_arr_int,0.0,0.0,0.0))
    
    #Dummy for sample_j_analytic
    if not bool(beam_models_list_type1):
        beam_models_list_type1.append((1,"none",0.,0.,0.0))
    
    return tuple(beam_models_list_type0), tuple(beam_models_list_type1), tuple(beam_types), tuple(new_beam_model_map)


def _exstract_arrays_from_bm_xds(bm):
    bm_J = bm.J.values 
    #print("bm_J", bm_J.shape)
    pa = bm.pa.values
    chan = bm.chan.values
    pol = bm.pol.values
    delta_l = (bm.l[1].values - bm.l[0].values).astype(float)
    delta_m = (bm.m[1].values - bm.m[0].values).astype(float)
    max_rad_1GHz = bm.attrs['max_rad_1GHz']
    return (0,bm_J,pa,chan,pol,delta_l,delta_m,max_rad_1GHz)
    
def _exstract_vals_from_analytic_dict(bm):
    pb_func = bm['pb_func']
    #dish_diameter = bm['dish_diameter']
    #blockage_diameter = bm['blockage_diameter']
    dish_diameter = bm['dish_diam']
    blockage_diameter = bm['blockage_diam']
    max_rad_1GHz = bm['max_rad_1GHz']
    return (1,pb_func,dish_diameter,blockage_diameter,max_rad_1GHz)


#@jit(nopython=True,cache=True,nogil=True)
@jit(nopython=True,nogil=True)
def _calc_pb_scale(flux, sep1, sep2, bm1_indx,bm2_indx,bm1_type,bm2_type,lmn1,lmn2,beam_models_type0,beam_models_type1,pa,freq,mueller_selection,do_pointing):

    #print(sep1,sep2,bm1_indx,bm2_indx,bm1_type,bm2_type)
    outside_beam = False
    if (bm1_indx == bm2_indx) and (bm1_type == bm2_type) and ~do_pointing: #Antennas are the same
        if bm1_type == 0: #Checks if it is a zernike model
            bm1 = beam_models_type0[bm1_indx]
            max_rad = bm1[7]/(freq/10**9) # scale max_rad_1GHz to freq
            
            if sep1 < max_rad:
                #bm1 = beam_models_type0[bm1_indx]
                J_sampled = _sample_J(bm1[1],bm1[2],bm1[3],bm1[4],bm1[5],bm1[6],pa,freq,lmn1)[:,0]
                
                #J_sampled = np.zeros((4,),dtype=numba.complex128)
                M = _make_mueler_mat(J_sampled, J_sampled, mueller_selection)
                #Add check if J sampled is < 0 and then skip this
                flux_scaled = np.dot(M,flux)
            else:
                outside_beam = True
        else:
            #bm1 = beam_models_type1[bm1_indx]
            bm1 = beam_models_type1[bm1_indx]
            max_rad = bm1[4]/(freq/10**9) # scale max_rad_1GHz to freq
            if sep1 < max_rad:
                J_sampled = _sample_J_analytic(bm1[1],bm1[2],bm1[3],max_rad,lmn1,freq,1)
                #print('J_sampled',J_sampled,bm1[1],bm1[2],bm1[3],lmn1,freq,1)
                flux_scaled = flux*J_sampled**2
            else:
                outside_beam = True
    else:
        if bm1_type == 0:
            bm1 = beam_models_type0[bm1_indx]
            max_rad = bm1[7]/(freq/10**9) # scale max_rad_1GHz to freq
            if sep1 < max_rad:
                J_sampled1 = _sample_J(bm1[1],bm1[2],bm1[3],bm1[4],bm1[5],bm1[6],pa,freq,lmn1)[:,0]
            else:
                outside_beam = True
        else:
            bm1 = beam_models_type1[bm1_indx]
            max_rad = bm1[4]/(freq/10**9) # scale max_rad_1GHz to freq
            if sep1 < max_rad:
                J_sampled1 = _sample_J_analytic(bm1[1],bm1[2],bm1[3],max_rad,lmn1,freq,1)
            else:
                outside_beam = True
        if bm2_type == 0:
            bm2 = beam_models_type0[bm2_indx]
            max_rad = bm2[7]/(freq/10**9) # scale max_rad_1GHz to freq
            if sep2 < max_rad:
                J_sampled2 = _sample_J(bm2[1],bm2[2],bm2[3],bm2[4],bm2[5],bm2[6],pa,freq,lmn2)[:,0]
            else:
                outside_beam =  True
        else:
            bm2 = beam_models_type1[bm2_indx]
            max_rad = bm2[4]/(freq/10**9) # scale max_rad_1GHz to freq
            if sep2 < max_rad:
                J_sampled2 = _sample_J_analytic(bm2[1],bm2[2],bm2[3],max_rad,lmn2,freq,1)
            else:
                outside_beam = True
                
        if not outside_beam:
            M = _make_mueler_mat(J_sampled1, J_sampled2, mueller_selection)
            flux_scaled = np.dot(M,flux)
    if outside_beam:
        flux_scaled = np.zeros(4, dtype = numba.complex128)

    return flux_scaled, outside_beam

@jit(nopython=True,cache=True,nogil=True)
def _pol_code_to_index(pol):
    if pol[0] in [5,6,7,8]:
        return pol-5
    if pol[0] in [9,10,11,12]:
        return pol-9
    assert False # "Unsupported pol " + str(pol) #Numba does not support an explicit error message.
    
#@jit(nopython=True,cache=True,nogil=True)
@jit(nopython=True,nogil=True)
def _sample_J_analytic(pb_func, dish_diameter,blockage_diameter,max_rad, lmn, freq, ipower):
    #pb_parms = bm
    #pb_parms['ipower'] = 1
    
    if pb_func == 'casa_airy':
        J_sampled = _casa_airy_pb(lmn,freq,dish_diameter, blockage_diameter,ipower,max_rad)
        #J_sampled = 0.5
    elif pb_func == 'airy':
        #J_sampled = 0.5
        J_sampled = _airy_pb(lmn,freq,dish_diameter, blockage_diameter, ipower)
    else:
        J_sampled = 1
    
    J_sampled_array = np.zeros(4, dtype = numba.float64)
    J_sampled_array[0] = J_sampled
    J_sampled_array[3] = J_sampled
    #J_sampled = np.array([J_sampled,0,0,J_sampled])
    return J_sampled_array.astype(numba.complex128)
    
@jit(nopython=True,cache=True,nogil=True)
def _sample_J(bm_J,bm_pa,bm_chan, bm_pol, bm_delta_l,bm_delta_m,pa,freq,lmn):
    pa_indx = _find_angle_indx(bm_pa,pa)
    chan_indx = _find_val_indx(bm_chan, freq)
    bm_J_sub = bm_J[pa_indx, chan_indx]

    x_rot, y_rot  = _rot_coord(lmn[0],lmn[1],pa-bm_pa[pa_indx])
    xrot = np.ones(1, dtype = numba.float64)
    xrot[0] = x_rot
    yrot = np.ones(1, dtype = numba.float64)
    yrot[0] = y_rot
    #print((xrot/bm[4]) + len(bm_J_sub[0, :, 0])//2)
    #print((yrot/bm[5]) + len(bm_J_sub[0, 0, :])//2)
    
    J_temp = _interp_array(bm_J_sub, xrot, yrot, bm_delta_l, bm_delta_m)
    
    J = np.zeros((4,J_temp.shape[1]),dtype=numba.complex128) + 1
    for i,p in enumerate(bm_pol):
        J[p] = J_temp[i,:]
        
    return J

#Non-numba versions:
def _sample_J_analytic_og(bm,lmn,freq):
    pb_parms = bm
    pb_parms['ipower'] = 1
    
    if pb_parms['pb_func'] == 'casa_airy':
        J_sampled = _casa_airy_pb_njit(lmn,freq,pb_parms)
    elif pb_parms['pb_func'] == 'airy':
        J_sampled = _airy_pb_njit(lmn,freq,pb_parms)
    else:
        J_sampled = 1
    J_sampled = np.array([J_sampled,0,0,J_sampled])
    return J_sampled


def _sample_J_og(bm,lmn,freq,pa):
# sample_J(J,pa_,chan_,lmn,chan,pa)
    bm_sub = bm
    if len(bm.pa) > 1:
        pa_indx= _find_angle_indx(bm_sub.pa.values,pa)
        bm_sub = bm_sub.isel(pa=pa_indx)
    else:
        bm_sub = bm_sub.isel(pa=0)
    if len(bm.chan) > 1:
        bm_sub = bm.J.interp(chan=freq,method='nearest')
    else:
        bm_sub = bm_sub.isel(chan=0)
        
    #print('pa values',bm_sub)
    x_rot, y_rot  = _rot_coord(lmn[0],lmn[1],pa-bm_sub.pa.values)
    
#    print(bm_sub)
#    print(lmn,pa,bm_sub.pa.values )
#    print(x_rot,y_rot)
#    print(bm_sub.J.isel(model=0).interp(l=x_rot,m=y_rot,method='linear'))
#    print(bm_sub.J.interp(l=x_rot,m=y_rot,method='linear').values)
    
    return bm_sub.J.interp(l=x_rot,m=y_rot,method='linear').values[0]

@jit(nopython=True,cache=True,nogil=True)
def _make_mueler_mat(J1, J2, mueller_selection):

    #M = np.zeros((4,4),dtype=np.complex)
    M = np.zeros((4,4),dtype=numba.complex128)
    
    for m_flat_indx in mueller_selection:
        #print(m_flat_indx//4,m_flat_indx - 4*(m_flat_indx//4))
        #print(np.where(map_mueler_to_pol[m_flat_indx][0] == pol)[0][0])
        #print(pol, map_mueler_to_pol[m_flat_indx][0])
        #M[m_flat_indx//4,m_flat_indx - 4*(m_flat_indx//4)] = J1[np.where(map_mueler_to_pol[m_flat_indx][0] == pol)[0][0]]*np.conj(J2[np.where(map_mueler_to_pol[m_flat_indx][1] == pol)[0][0]])
        
        M[m_flat_indx//4,m_flat_indx - 4*(m_flat_indx//4)] = J1[map_mueler_to_pol[m_flat_indx,0]]*np.conj(J2[map_mueler_to_pol[m_flat_indx,1]])
            
    return M
    
    #map_mueler_to_pol = np.array([[0,0],[0,1],[1,0],[1,1],[0,2],[0,3],[1,2],[1,3],[2,0],[2,1],[3,0],[3,1],[2,2],[2,3],[3,2],[3,3]])

'''
    #print('pol',pol)
    if inv:
        map_mueler_to_pol = np.array([[3, 3],[3, 2],[2, 3],[2, 2],[3, 1],[3, 0],[2, 1],[2, 0],[1, 3],[1, 2],[0, 3],[0, 2],[1, 1],[1, 0],[0, 1],[0, 0]]) # np.flip(map_mueler_to_pol,axis=0)
        #map_mueler_to_pol = np.array([ [[3, 3],[3, 2],[2, 3],[2, 2]],[[3, 1],[3, 0],[2, 1],[2, 0]],[[1, 3],[1, 2],[0, 3],[0, 2]],[[1, 1],[1, 0],[0, 1],[0, 0]]])
    else:
        map_mueler_to_pol = np.array([[0,0],[0,1],[1,0],[1,1],[0,2],[0,3],[1,2],[1,3],[2,0],[2,1],[3,0],[3,1],[2,2],[2,3],[3,2],[3,3]])
        #map_mueler_to_pol = np.array([[[0,0],[0,1],[1,0],[1,1]],[[0,2],[0,3],[1,2],[1,3]],[[2,0],[2,1],[3,0],[3,1]],[[2,2],[2,3],[3,2],[3,3]]])
        
'''

# Do we need the abs for apeture_parms['zernike_size'].
#    apeture_parms['cell_size'] = 1/(beam_parms['cell_size']*beam_parms['image_size']*apeture_parms['oversampling'])
#    apeture_parms['zernike_size'] = np.floor(np.abs(Dish_Diameter*eta/(apeture_parms['cell_size']*lmbd))) #Why is abs used?
#2.) Currently in ARD-20 the Zernike grid parameters are caculated using (in ZernikeCalc.cc):
#size_zernike = floor((D*eta)/dx)
#delta_zernike = 2.0/size_zernike
#
#By reversing the order the delta of the zernike grid and uv grid can match exactly (no floor operation). The zernike grid will be slightly larger (all cells outside the r=1 will still be nulled).
#delta_zernike = (2.0*dx)/(D*eta)
#size_zernike = ceil((D*eta)/dx)
#Assume ETA is the same for all pol and coef. Only a function of freq. Need to update zpc.zarr format.
  
def _calc_ant_jones(zpc_dataset,j_freq,j_pa,beam_parms):
    pa_prev = -42.0
    freq_prev = -42.0
    
    j_planes = np.zeros((len(j_pa),len(j_freq),len(beam_parms['needed_pol']),beam_parms['image_size'][0],beam_parms['image_size'][1]),np.complex128)
    j_planes_shape = j_planes.shape
    iter_dims_indx = itertools.product(np.arange(j_planes_shape[0]),np.arange(j_planes_shape[1]))
    ic = beam_parms['image_size']//2 #image center pixel
    
    for i_pa, i_chan in iter_dims_indx:
        #print(i_pa,i_chan)
        pa = j_pa[i_pa]
        beam = zpc_dataset
        freq = j_freq[i_chan]

        if(freq != freq_prev):
            beam_interp = beam.interp(chan=freq,method=beam_parms['zernike_freq_interp'])
            
        dish_diam = beam.dish_diam
        lmbd = c/freq
        eta = beam_interp.ETA[0,0].values #Assume ETA is the same for all pol and coef. Only a function of freq. Need to update zpc.zarr format.
        uv_cell_size = 1/(beam_parms['cell_size']*beam_parms['image_size'])
        zernike_cell = (2.0*uv_cell_size*lmbd)/(dish_diam*eta)
        
        
        if (pa != pa_prev) or (freq != freq_prev) :
            beam_parms['parallactic_angle'] = pa
            image_size = (np.ceil(np.abs(2.0/zernike_cell))).astype(int)
            x_grid, y_grid = _compute_rot_coords(image_size,zernike_cell,pa)
            
            r_grid = np.sqrt(x_grid**2 + y_grid**2)
            
            zernike_size = np.array(x_grid.shape)
        
            ic_z = zernike_size//2
            include_last = (zernike_size%2).astype(int)
        
        #assert zernike_size[0] < beam_parms['conv_size'][0] and zernike_size[1] < gcf_parms['conv_size'][1], "The convolution size " + str(gcf_parms['conv_size']) +" is smaller than the aperture image " + zernike_size + " . Increase conv_size"
        
        #print('x_grid',x_grid)
        #print('y_grid',y_grid)
        
        
        start = time.time()
        for i_pol,pol in enumerate(beam_parms['needed_pol']):
            a = _generate_zernike_surface(beam_interp.ZC.data[pol,:],x_grid,y_grid)
            a[r_grid > 1] = 0
            j_planes[i_pa, i_chan,i_pol,ic[0]-ic_z[0]:ic[0]+ic_z[0]+include_last[0],ic[1]-ic_z[1]:ic[1]+ic_z[1]+include_last[1]] = a
            j_planes[i_pa, i_chan, i_pol,:,:] = np.fft.fftshift(np.fft.ifft2(np.fft.ifftshift(j_planes[i_pa, i_chan,i_pol,:,:])))/(beam_parms['image_size'][0]*beam_parms['image_size'][1])
        #print('One pol set',time.time()-start)
        
        #Normalize Jones
        if 3 not in beam_parms['needed_pol']:
            P_max = np.abs(j_planes[i_pa, i_chan, np.where(beam_parms['needed_pol']==0),j_planes_shape[3]//2,j_planes_shape[4]//2])
            Q_max = P_max
        elif 0 not in beam_parms['needed_pol']:
            Q_max = np.abs(j_planes[i_pa, i_chan, np.where(beam_parms['needed_pol']==3),j_planes_shape[3]//2,j_planes_shape[4]//2])
            P_max = Q_max
        else:
            P_max = np.abs(j_planes[i_pa, i_chan, np.where(beam_parms['needed_pol']==0),j_planes_shape[3]//2,j_planes_shape[4]//2])
            Q_max = np.abs(j_planes[i_pa, i_chan, np.where(beam_parms['needed_pol']==3),j_planes_shape[3]//2,j_planes_shape[4]//2])

        j_planes[i_pa, i_chan,:,:,:] = j_planes[i_pa, i_chan,:,:,:]*2/(P_max+Q_max)
        
        pa_prev = pa
        freq_prev = freq
    return j_planes#np.zeros((1,4,2048,2048),dtype=np.complex128)



import numpy as np
from sirius.calc_vis import sample_J, sample_J_analytic
from sirius._sirius_utils._direction_rotate import _directional_cosine

def test_sample_J():
    x = np.arange(0, 201, 1)
    y = np.arange(0, 201, 1)
    xx, yy = np.meshgrid(x, y)
    z = np.sin(xx**2+(yy+1)**2)
    bm_J = np.array([[[z, z], [z, z], [z, z]], [[z, z], [z, z], [z, z]]])
    bm_pa = np.array([0, 1])
    bm_chan = np.array([0, 1, 2])
    bm_pol = np.array([0, 1, 2])
    lmn = _directional_cosine((2.1, 3.2))
    freq = 1.1
    pa = 0.8
    delta_l = 4
    delta_m = 4
    test1 = np.array([[0.529615457+0.j],[0.529615457+0.j], [0.+0.j], [1.+0.j]], dtype="complex128")
    test2 = sample_J(bm_J, bm_pa, bm_chan, bm_pol, delta_l, delta_m, pa, freq, lmn)
    assert np.allclose(test1, test2, rtol = 1e-8) == True
    
    
def test_sample_J_analytic_airy():
    lmn = _directional_cosine((2.1, 3.2))
    assert np.allclose(np.array([-0.00019941+0.j, 0.+0.j, 0.+0.j, -0.00019941+0.j]), sample_J_analytic("airy", 25, 2, lmn, 1.2e9, 1)) == True
    
def test_sample_J_analytic_CASA():
    lmn = _directional_cosine((2.1, 3.2))
    #assert np.allclose(np.array([-0.00025785+0.j, 0.+0.j, 0.+0.j, -0.00025785+0.j]), sample_J_analytic("casa_airy", 25, 2, lmn, 1.2e9, 1)) == True
    assert np.allclose(np.array([-0.00025785+0.j,  0.        +0.j,  0.        +0.j, -0.00025785+0.j]), sample_J_analytic("casa_airy", 25, 2, lmn, 1.2e9, 1), rtol = 1e-8) == True


 #   Copyright 2019 AUI, Inc. Washington DC, USA
 #
 #   Licensed under the Apache License, Version 2.0 (the "License");
 #   you may not use this file except in compliance with the License.
 #   You may obtain a copy of the License at
 #
 #       http://www.apache.org/licenses/LICENSE-2.0
 #
 #   Unless required by applicable law or agreed to in writing, software
 #   distributed under the License is distributed on an "AS IS" BASIS,
 #   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 #   See the License for the specific language governing permissions and
 #   limitations under the License.
 
import numpy as np
import xarray as xr
from astropy.timeseries import TimeSeries
from astropy.time import Time
from astropy import units as u
import dask.array as da
from collections import Counter
import time
import dask
import os
 
def make_time_xda(time_start='2019-10-03T19:00:00.000',time_delta=3600,n_samples=10,n_chunks=4):
    """
    Create a time series xarray array.
    Parameters
    ----------
    -------
    time_da : dask.array
    """
    ts = np.array(TimeSeries(time_start=time_start,time_delta=time_delta*u.s,n_samples=n_samples).time.value)
    chunksize = int(np.ceil(n_samples/n_chunks))
    time_da = da.from_array(ts, chunks=chunksize)
    print('Number of chunks ', len(time_da.chunks[0]))
    
    time_xda = xr.DataArray(data=time_da,dims=["time"],attrs={'time_delta':float(time_delta)})
    
    return time_xda

def make_chan_xda(spw_name='sband',freq_start = 3*10**9, freq_delta = 0.4*10**9, freq_resolution=0.01*10**9, n_channels=3, n_chunks=3):
    """
    Create a channel frequencies xarray array.
    Parameters
    ----------
    -------
    time_da : dask.array
    """
    freq_chan = (np.arange(0,n_channels)*freq_delta + freq_start).astype(float) #astype(float) needed for interfacing with CASA simulator.
    chunksize = int(np.ceil(n_channels/n_chunks))
    chan_da = da.from_array(freq_chan, chunks=chunksize)
    print('Number of chunks ', len(chan_da.chunks[0]))
    
    chan_xda = xr.DataArray(data=chan_da,dims=["chan"],attrs={'freq_resolution':float(freq_resolution),'spw_name':spw_name, 'freq_delta':float(freq_delta)})
    return chan_xda
      
def write_to_ms(vis_xds, time_xda, chan_xda, pol, tel_xds, phase_center_names, phase_center_ra_dec, auto_corr,save_parms):
    
    if save_parms['write_to_ms']:
        start = time.time()
        from casatools import simulator
        from casatasks import mstransform
        n_time, n_baseline, n_chan, n_pol = vis_xds.DATA.shape

        sm = simulator()
        
        ant_pos = tel_xds.ANT_POS.values
        os.system('rm -rf ' + save_parms['ms_name'])
        sm.open(ms=save_parms['ms_name']);

        ###########################################################################################################################
        ## Set the antenna configuration
        sm.setconfig(telescopename= tel_xds.telescope_name,
                        x=ant_pos[:,0],
                        y=ant_pos[:,1],
                        z=ant_pos[:,2],
                        dishdiameter=tel_xds.DISH_DIAMETER.values,
                        mount=['alt-az'],
                        antname=list(tel_xds.ant_name.values),  #CASA can't handle an array of antenna names.
                        coordsystem='global',
                        referencelocation=tel_xds.site_pos[0]);
                        
        ## Set the polarization mode (this goes to the FEED subtable)
        from sirius_data._constants import pol_codes_RL, pol_codes_XY, pol_str
        from sirius._sirius_utils._array_utils import _is_subset
        if _is_subset(pol_codes_RL,pol): #['RR','RL','LR','LL']
            sm.setfeed(mode='perfect R L', pol=['']);
        elif _is_subset(pol_codes_XY,pol): #['XX','XY','YX','YY']
            sm.setfeed(mode='perfect X Y', pol=['']);
        else:
            assert False, print('Pol selection invalid, must either be subset of [5,6,7,8] or [9,10,11,12] but is ', pol)

        sm.setspwindow(spwname=chan_xda.spw_name,
                    freq=chan_xda.data[0].compute(),
                    deltafreq=chan_xda.freq_delta,
                    freqresolution=chan_xda.freq_resolution,
                    nchannels=len(chan_xda),
                    refcode='LSRK',
                    stokes=' '.join(pol_str[pol]));


        if auto_corr:
            sm.setauto(autocorrwt=1.0)
        else:
            sm.setauto(autocorrwt=0.0)
            
        mjd = Time(time_xda.data[0:2].compute(), scale='utc')
        integration_time = (mjd[1]-mjd[0]).to('second')

        start_time = (mjd[0] - (integration_time/2 + 37*u.second)).mjd
        start_time_dict = {'m0': {'unit': 'd', 'value': start_time}, 'refer': 'UTC', 'type': 'epoch'}

        sm.settimes(integrationtime=integration_time.value,
                    usehourangle=False,
                    referencetime=start_time_dict);
            
        fields_set = []
        field_time_count = Counter(phase_center_names)

        #print(field_time_count,phase_center_names)
        if len(phase_center_names) == 1: #Single field case
            field_time_count[list(field_time_count.keys())[0]] = n_time

        start_time = 0
        for i,ra_dec in enumerate(phase_center_ra_dec): #In future make phase_center_ra_dec a unique list
            if phase_center_names[i] not in fields_set:
                dir_dict = {'m0': {'unit': 'rad', 'value': ra_dec[0]}, 'm1': {'unit': 'rad', 'value': ra_dec[1]}, 'refer': 'J2000', 'type': 'direction'}
                sm.setfield(sourcename=phase_center_names[i],sourcedirection=dir_dict)
                fields_set.append(phase_center_names[i])
                
                stop_time = start_time + integration_time.value*field_time_count[phase_center_names[i]]
                sm.observe(sourcename=phase_center_names[i],
                    spwname=chan_xda.spw_name,
                    starttime= str(start_time) + 's',
                    stoptime= str(stop_time) + 's')
                start_time = stop_time
                

        n_row = n_time*n_baseline

        print('Meta data creation ',time.time()-start)

        #print(vis_data.shape)
        #print(n_row,n_time, n_baseline, n_chan, n_pol)

        start = time.time()
        #This code will most probably be moved into simulation if we get rid of row time baseline split.
        vis_data_reshaped = vis_xds.DATA.data.reshape((n_row, n_chan, n_pol))
        uvw_reshaped = vis_xds.UVW.data.reshape((n_row, 3))
        weight_reshaped = vis_xds.WEIGHT.data.reshape((n_row,n_pol))
        sigma_reshaped = vis_xds.SIGMA.data.reshape((n_row,n_pol))

        print('reshape time ', time.time()-start)
        #weight_spectrum_reshaped = np.tile(weight_reshaped[:,None,:],(1,n_chan,1))



        #    print(weight_reshaped.compute().shape)
        #    print(sigma_reshaped.compute().shape)
        #    print(weight_reshaped)
        #    print(sigma_reshaped)

        #dask_ddid = da.full(n_row, 0, chunks=chunks['row'], dtype=np.int32)

        #print('vis_data_reshaped',vis_data_reshaped)

        start = time.time()
        from daskms import xds_to_table, xds_from_ms, Dataset

        #print('vis_data_reshaped.chunks',vis_data_reshaped.chunks)
        row_id = da.arange(n_row,chunks=vis_data_reshaped.chunks[0],dtype='int32')


        dataset = Dataset({'DATA': (("row", "chan", "corr"), vis_data_reshaped), 'CORRECTED_DATA': (("row", "chan", "corr"), vis_data_reshaped),'UVW': (("row","uvw"), uvw_reshaped), 'SIGMA': (("row","pol"), sigma_reshaped), 'WEIGHT': (("row","pol"), weight_reshaped),  'ROWID': (("row",),row_id)})
        #,'WEIGHT_SPECTRUM': (("row","chan","pol"), weight_spectrum_reshaped)
        ms_writes = xds_to_table(dataset, save_parms['ms_name'], columns="ALL")

        if save_parms['DAG_name_write']:
            dask.visualize(ms_writes,filename=save_parms['DAG_name_write'])
            
        if save_parms['write_to_ms']:
            start = time.time()
            dask.compute(ms_writes)
            print('*** Dask compute time',time.time()-start)

        print('compute and save time ', time.time()-start)

        sm.close()

        from casatasks import flagdata
        flagdata(vis=save_parms['ms_name'],mode='unflag')
        
        
        return xds_from_ms(save_parms['ms_name'])




import numpy as np
from scipy.interpolate import interp2d
from sirius._sirius_utils._math_utils import _bilinear_interpolate, _interp_array

def test_2d_interpolation():
    x = np.arange(0, 201, 1)
    y = np.arange(0, 201, 1)
    xx, yy = np.meshgrid(x, y)
    z = np.sin(xx**2+(yy+1)**2)
    f = interp2d(x, y, z, kind='linear')
    
    assert np.allclose(f(45.5, 51.5), _bilinear_interpolate(z, np.array([45.5]), np.array([51.5]))) == True
    
def test_array_interpolation():
    x = np.arange(0, 201, 1)
    y = np.arange(0, 201, 1)
    xx, yy = np.meshgrid(x, y)
    z = np.sin(xx**2+(yy+1)**2)
    z_c = np.sin(xx**2+(yy+1)**2)*0
    assert np.allclose(np.array([[0.56429664+0.j], [1.12859327+0.j], [1.69288991+0.j]]), _interp_array(np.array([z+1j*z_c, 2*z+1j*z_c, 3*z+1j*z_c]), np.array([2.]), np.array([2.]), 4., 4.)) == True
    
def test_array_interpolation_complex():
    x = np.arange(0, 201, 1)
    y = np.arange(0, 201, 1)
    xx, yy = np.meshgrid(x, y)
    z = np.sin(xx**2+(yy+1)**2)
    assert np.allclose(np.array([[0.56429666+1.1285933j ], [0.56429666+0.56429666j], [0.56429666+0.j]], dtype="complex128"), _interp_array(np.array([z+2j*z, z+1j*z, z]), np.array([2.]), np.array([2.]), 4., 4.)) == True


#   Copyright 2019 AUI, Inc. Washington DC, USA
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from scipy.spatial.transform import Rotation as R
import numpy as np
from numba import jit, types
import numba


@jit(nopython=True,cache=True,nogil=True)
def _func_R_z(phi):
    result = np.zeros((3, 3), dtype = numba.float64)
    result[0, 0] = np.cos(phi)
    result[0, 1] = -np.sin(phi)
    result[1, 0] = np.sin(phi)
    result[1, 1] = np.cos(phi)
    result[2, 2] = 1
    return result

#def _func_R_z(phi):
#    return np.array([[np.cos(phi), -np.sin(phi), 0],[np.sin(phi), np.cos(phi), 0],[0, 0, 1]])

@jit(nopython=True,cache=True,nogil=True)
def _func_R_x(theta):
    result = np.zeros((3, 3), dtype = numba.float64)
    result[0, 0] = 1
    result[1, 1] = np.cos(theta)
    result[1, 2] = -np.sin(theta)
    result[2, 1] = np.sin(theta)
    result[2, 2] = np.cos(theta)
    return result

#def _func_R_x(theta):
#    return np.array([[1, 0, 0],[0, np.cos(theta), -np.sin(theta)],[0, np.sin(theta), np.cos(theta)]])

@jit(nopython=True,cache=True,nogil=True)
def _func_R_y(psi):
    result = np.zeros((3, 3), dtype = numba.float64)
    result[0, 0] = np.cos(psi)
    result[0, 2] = np.sin(psi)
    result[1, 1] = 1
    result[2, 0] = -np.sin(psi)
    result[2, 2] = np.cos(psi)
    return result
    
#def _func_R_y(psi):
#    return np.array([[np.cos(psi), 0, np.sin(psi)],[0, 1, 0],[-np.sin(psi), 0, np.cos(psi)]])

@jit(nopython=True,cache=True,nogil=True)
def _directional_cosine(ra_dec):
    #The coordinate definition of the measurement set V2/3 is used (https://casacore.github.io/casacore-notes/264.pdf see page 12 and https://drive.google.com/file/d/1a-eUwNrfnYjaUQTjJDfOjJCa8ZaSzZcn/view?usp=sharing).
    #Note this does not follow the convention in chapter 4 of https://link.springer.com/book/10.1007/978-3-319-44431-4 or CASACORE.
    #Celestial coordinate system is defined by:
    #        Z pointing to NCP (dec = pi/2),
    #        X pointing to the East (ra = pi/2, dec = 0),
    #       -Y pointing to the vernal equinox (ra = 0, dec = 0).
    # The uvw coordinates are aligned with XYZ when the source is located at the NCP (ra=0 and dec=pi/2).
    # A good overview of coordinates systems [1] http://www.danfleisch.com/maxwell/CoordinateSystemReview.pdf
    # Using the physics convention for spherical coordinates:
    #        Polar angle      theta = pi/2 - dec (1)
    #        Azimuthal Angle  phi   = 3*pi/2 + ra (2)
    #Substituting (1) and (2) into equations (8)-(10) of [1] yields:
    lmn = np.zeros((3,))
    lmn[0] = np.sin(ra_dec[0])*np.cos(ra_dec[1])
    lmn[1] = -np.cos(ra_dec[0])*np.cos(ra_dec[1])
    lmn[2] = np.sin(ra_dec[1])
   
    return lmn
    
@jit(nopython=True,cache=True,nogil=True)
def _sin_project(ra_dec_o,ra_dec):
    #Orthographic/Synthesis projection of right ascension and declination to a tangential plane with center at ra_dec_o.
    #Equations 10 http://tdc-www.harvard.edu/wcstools/aips27.pdf
    ra = ra_dec[0]
    dec = ra_dec[1]
    ra_o = ra_dec_o[0]
    dec_o = ra_dec_o[1]
    lm = np.zeros((2,),dtype = numba.float64)
    lm[0] = np.cos(dec)*np.sin(ra-ra_o)
    lm[1] = np.sin(dec)*np.cos(dec_o) - np.cos(dec)*np.sin(dec_o)*np.cos(ra-ra_o)
    return lm
    
    
@jit(nopython=True,cache=True,nogil=True)
def _tan_project(ra_dec_o,ra_dec):
    #Gnomonic projection of right ascension and declination to a tangential plane with center at ra_dec_o.
    #Equations 9 http://tdc-www.harvard.edu/wcstools/aips27.pdf
    ra = ra_dec[0]
    dec = ra_dec[1]
    ra_o = ra_dec_o[0]
    dec_o = ra_dec_o[1]
    lm = np.zeros((2,),dtype = numba.float64)
    div_term = np.sin(dec)*np.sin(dec_o) + np.cos(dec)*np.cos(dec_o)*np.cos(ra-ra_o)
    lm[0] = np.cos(dec)*np.sin(ra-ra_o)/div_term
    lm[1] = (np.sin(dec)*np.cos(dec_o) - np.cos(dec)*np.sin(dec_o)*np.cos(ra-ra_o))/div_term
    return lm
    
@jit(nopython=True,cache=True,nogil=True)
def _local_spherical_coords(ra_dec_o,ra_dec):
    #Equation 6-8 http://tdc-www.harvard.edu/wcstools/aips27.pdf
    ra = ra_dec[0]
    dec = ra_dec[1]
    ra_o = ra_dec_o[0]
    dec_o = ra_dec_o[1]
    theta_phi = np.zeros((2,),dtype = numba.float64)
    theta_phi[0] = np.arccos(np.sin(dec)*np.sin(dec_o) + np.cos(dec)*np.cos(dec_o)*np.cos(ra-ra_o))
    theta_phi[1] = np.arctan2(np.cos(dec)*np.sin(ra-ra_o),np.sin(dec)*np.cos(dec_o) + np.cos(dec)*np.cos(dec_o)*np.cos(ra-ra_o))
    
    return theta_phi
    

@jit(nopython=True,cache=True,nogil=True)
def _calc_rotation_mats(ra_dec_in,ra_dec_out):
    
    #The default
    #Rotates input system to output system.
    uvw_rotmat = _func_R_x(-(np.pi/2-ra_dec_in[1]))@_func_R_z(ra_dec_out[0]-ra_dec_in[0])@_func_R_x(np.pi/2 - ra_dec_out[1])
    
    #Rotates output to Standard System
    out_rotmat = _func_R_x(-(np.pi/2-ra_dec_out[1]))@_func_R_z(-ra_dec_out[0])
    lmn_out = _directional_cosine(ra_dec_out)
    lmn_in = _directional_cosine(ra_dec_in)
    
    lmn_rot = out_rotmat@(lmn_out - lmn_in) # out_rotmat(lmn_out - lmn_in)= [0,0,1] - out_rotmat@[l_in,m_in,n_in]
    
    return uvw_rotmat, lmn_rot


@jit(types.containers.UniTuple(numba.float64, 2)(numba.float64, numba.float64, numba.float64), nopython=True,cache=True,nogil=True)
def _rot_coord(x,y,parallactic_angle):
    x_rot = np.cos(parallactic_angle)*x + np.sin(parallactic_angle)*y
    y_rot = - np.sin(parallactic_angle)*x + np.cos(parallactic_angle)*y
    return x_rot,y_rot


def _compute_rot_coords(image_size,cell_size,parallactic_angle):
    image_center = image_size//2
    #print(image_size)

    x = np.arange(-image_center[0], image_size[0]-image_center[0])*cell_size[0]
    y = np.arange(-image_center[1], image_size[1]-image_center[1])*cell_size[1]
    #xy = np.array([x,y]).T
    x_grid, y_grid = np.meshgrid(x,y,indexing='ij')
    
    if parallactic_angle != 0:
        #rot_mat = np.array([[np.cos(parallactic_angle),-np.sin(parallactic_angle)],[np.sin(parallactic_angle),np.cos(parallactic_angle)]]) #anti clockwise
        
        #r = np.einsum('ji, mni -> jmn', rot_mat, np.dstack([x_grid, y_grid]))
        '''
        x_grid_rot = np.cos(parallactic_angle)*x_grid - np.sin(parallactic_angle)*y_grid
        y_grid_rot = np.sin(parallactic_angle)*x_grid + np.cos(parallactic_angle)*y_grid
        '''
        x_grid_rot = np.cos(parallactic_angle)*x_grid + np.sin(parallactic_angle)*y_grid
        y_grid_rot = - np.sin(parallactic_angle)*x_grid + np.cos(parallactic_angle)*y_grid
        
        x_grid = x_grid_rot
        y_grid = y_grid_rot
    
    return x_grid, y_grid








################################################
#CASACORE derived functions, not used

#cs functions: casa versions.
def _cs_directional_cosine(ra_dec):
    # In https://arxiv.org/pdf/astro-ph/0207413.pdf see equation 160
    #ra_dec (RA,DEC)
    #lmn = np.zeros((3,),dtype=numba.f8)
    lmn = np.zeros((3,))
    lmn[0] = np.cos(ra_dec[0])*np.cos(ra_dec[1])
    lmn[1] = np.sin(ra_dec[0])*np.cos(ra_dec[1])
    lmn[2] = np.sin(ra_dec[1])
    return lmn


def _cs_calc_rotation_mats(ra_dec_in,ra_dec_out,rotation_parms):
    #ra_dec_in
    #ra_dec_out
    #rotation_parms
    #'common_tangent_reprojection'
    #'reproject'
    #https://github.com/casacore/casacore/blob/dbf28794ef446bbf4e6150653dbe404379a3c429/measures/Measures/UVWMachine.cc
    #https://github.com/casacore/casacore/blob/dbf28794ef446bbf4e6150653dbe404379a3c429/measures/Measures/UVWMachine.h


    #The rotation matrix from a system that has a pole towards output
    #direction, into the standard system.
    out_rotmat = R.from_euler('XZ',[[- ra_dec_out[1] + np.pi/2, - ra_dec_out[0] + np.pi/2]]).as_matrix()[0]
    #print('rot3_p',out_rotmat)
    out_dir_cosine = _cs_directional_cosine(ra_dec_out)
    
    uvw_rotmat = np.zeros((3,3),np.double)
    phase_rotation = np.zeros((3,),np.double)
    
    # Define rotation to a coordinate system with pole towards in-direction
    # and X-axis W; by rotating around z-axis over -(90-long); and around
    # x-axis (lat-90).
    in_rotmat = R.from_euler('ZX',[[ra_dec_in[0] - np.pi/2 , ra_dec_in[1] - np.pi/2]]).as_matrix()[0]
    
#    print(-np.pi/2, ra_dec_in[0])
#    aR_z = R.from_euler('Z',[[-np.pi/2 + ra_dec_in[0]]]).as_matrix()[0]
#    print('R_z',aR_z)
#
#    aR_x = R.from_euler('X',[[- np.pi/2 + ra_dec_in[1]]]).as_matrix()[0]
#    print('R_x ',aR_x)
#    #print('rot1_p',in_rotmat)
#    print('compare',in_rotmat-aR_z@aR_x)
    R_zx_si = _func_R_z(ra_dec_in[0] - np.pi/2)@_func_R_x(ra_dec_in[1] - np.pi/2) #in_rotmat
    R_xz_is = _func_R_x(-ra_dec_in[1] + np.pi/2)@_func_R_z(- ra_dec_in[0] + np.pi/2) # R_xz_is = R_zx_si.T
    print('@@@@@@@@@@@in_rotmat',mat_dis(R_zx_si,in_rotmat))
    
    R_xz_os = _func_R_x(- ra_dec_out[1] + np.pi/2)@_func_R_z(- ra_dec_out[0] + np.pi/2) #out_rotmat
    R_zx_so = _func_R_z(ra_dec_out[0] - np.pi/2)@_func_R_x(ra_dec_out[1] - np.pi/2)
    print('@@@@@@@@@@@out_rotmat',mat_dis(R_xz_os,out_rotmat))
    
    uvw_rotmat = np.matmul(out_rotmat,in_rotmat).T
    
    print('@@@@@@@@@@@uvw_rotmat',mat_dis(uvw_rotmat,(R_xz_os@R_zx_si ).T))
    print('@@@@@@@@@@@uvw_rotmat',mat_dis(uvw_rotmat,(R_xz_is@R_zx_so)))
    
    #print('uvrot_p',uvw_rotmat[:,:])
    uvw_proj_rotmat = None
    
    if rotation_parms['reproject'] == True:
        #Get the rotation matrix which re-projects an uv-plane onto another reference direction:
        # around x-axis (out-lat - 90)
        # around z-axis (out-long - in-long)
        # around x-axis (90 - in-lat) and normalise
        proj_out_rotmat = np.eye(3)
        temp = R.from_euler('XZX',[[-np.pi/2 + ra_dec_out[1], ra_dec_out[0] - ra_dec_in[0], np.pi/2 - ra_dec_in[1] ]]).as_matrix()[0]
        proj_out_rotmat[0,0] = temp[1,1]/temp[2,2]
        proj_out_rotmat[1,1] = temp[0,0]/temp[2,2]
        proj_out_rotmat[0,1] = temp[1,0]/temp[2,2]
        proj_out_rotmat[1,0] = temp[0,1]/temp[2,2]
        uvw_proj_rotmat = np.matmul(uvw_rotmat,proj_out_rotmat)
    #print('rot4_p',proj_out_rotmat)
    
    #print('uvproj_p',uvw_rotmat[:,:])
    
    if rotation_parms['common_tangent_reprojection'] == True:
        uvw_rotmat[2,0:2] = 0.0 # (Common tangent rotation needed for joint mosaics, see last part of FTMachine::girarUVW in CASA)
    
    in_dir_cosine = _cs_directional_cosine(ra_dec_in)
    #print("i_field, field, new",i_field,field_phase_center_cosine,new_phase_center_cosine)
    phase_rotation = np.matmul(out_rotmat,(out_dir_cosine - in_dir_cosine))
    
    print('in_rotmat',in_rotmat)
    print('out_rotmat',out_rotmat)
    print('out_dir_cosine',out_dir_cosine)
    print('in_dir_cosine',in_dir_cosine)
    print('phase_rotation',phase_rotation)
    print((out_rotmat@(out_dir_cosine - in_dir_cosine)).shape)
    print('out_rotmat,out_dir_cosine',np.matmul(out_rotmat,out_dir_cosine))
    print('out_rotmat,-in_dir_cosine',np.matmul(out_rotmat,-in_dir_cosine))
    
    #print('phrot_p',phase_rotation)
    
    return uvw_rotmat, uvw_proj_rotmat, phase_rotation




#  CASA Next Generation Infrastructure
#  Copyright (C) 2021 AUI, Inc. Washington DC, USA
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

#Get 1GHz max rad values from https://open-bitbucket.nrao.edu/projects/CASA/repos/casa6/browse/casatools/src/code/synthesis/TransformMachines/PBMath.cc
#Values is in degrees. Convert to radians.
import numpy as np

vla =  {'pb_func':'casa_airy','dish_diam':24.5,'blockage_diam':0.0,'max_rad_1GHz':0.8564*np.pi/180}
aca =  {'pb_func':'casa_airy','dish_diam':6.25,'blockage_diam':0.75,'max_rad_1GHz':3.568*np.pi/180}
alma =  {'pb_func':'casa_airy','dish_diam':10.7,'blockage_diam':0.75,'max_rad_1GHz':1.784*np.pi/180}
ngvla = {'pb_func':'casa_airy','dish_diam':18.0,'blockage_diam':0.0,'max_rad_1GHz':1.5*np.pi/180}


#  CASA Next Generation Infrastructure
#  Copyright (C) 2021 AUI, Inc. Washington DC, USA
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

#Do not modify
import numpy as np

map_mueler_to_pol = np.array([[0,0],[0,1],[1,0],[1,1],[0,2],[0,3],[1,2],[1,3],[2,0],[2,1],[3,0],[3,1],[2,2],[2,3],[3,2],[3,3]])

#https://github.com/casacore/casacore/blob/dbf28794ef446bbf4e6150653dbe404379a3c429/measures/Measures/Stokes.h
pol_codes_RL = np.array([5,6,7,8]) #'RR','RL','LR','LL'
pol_codes_XY = np.array([9,10,11,12]) #['XX','XY','YX','YY']
pol_str = np.array(['0','I','Q','U','V','RR','RL','LR','LL','XX','XY','YX','YY'])

k_B = 1.380649*10**-23 #Boltzmann constant J.K**-1
t_cmb = 2.725 #Cosmic Background Temperature
c = 299792458


'''
*****NB Note SiRIUS does not CGS units but MKS units****

https://open-bitbucket.nrao.edu/projects/CASA/repos/casa6/browse/casatools/src/python/constants.py
# %&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&
# IMPORTS
# %&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&

import numpy as _numpy

# %&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&
# CGS PHYSICAL CONSTANTS
# %&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&

c = 2.99792458e10       # speed of light CGS
h = 6.6260755e-27       # Planck's constant CGS
g = 6.67259e-8          # Grav const CGS
kb = 1.380658e-16       # Boltzmann's const CGS
a = 7.56591e-15         # Radiation constant CGS
sb = 5.67051e-5         # sigma (stefan-boltzmann const) CGS
qe =  4.803206e-10      # Charge of electron CGS
ev =  1.60217733e-12    # Electron volt CGS
na =  6.0221367e23      # Avagadro's Number
me =  9.1093897e-28     # electron mass CGS
mp =  1.6726231e-24     # proton mass CGS
mn = 1.674929e-24       # neutron mass CGS
mh = 1.673534e-24       # hydrogen mass CGS
amu =  1.6605402e-24    # atomic mass unit CGS

# %&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&
# ASTRONOMICAL CONSTANTS
# %&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&

# GENERAL
au = 1.496e13           # astronomical unit CGS
pc = 3.0857e18          # parsec CGS
yr = 3.155815e7         # sidereal year CGS
ms = 1.98900e+33        # solar mass CGS
rs = 6.9599e10          # sun's radius CGS
ls = 3.839e33           # sun's luminosity CGS
mm = 7.35000e+25        # moon mass CGS
mer = 5.97400e+27       # earth mass CGS
rer = 6.378e8           # earth's radius CGS
medd = 3.60271e+34      # Eddington mass CGS

# RADIO SPECIFIC
jy = 1.e-23                  # Jansky CGS
restfreq_hi = 1420405751.786 # 21cm transition (Hz)
restfreq_co = 115271201800.  # CO J=1-0 (Hz)
cm2perkkms_hi = 1.823e18     # HI column per intensity (thin)

# OTHER WAVELENGTHS
ksun = 3.28             # abs K mag of sun (Binney & Merrifield)

# %&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&
# GEOMETRY
# %&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&

srdeg = 3283.0          # Degrees squared in a steradian
dtor =  0.0174532925199 # Degrees per radian
pi = 3.14159265359      # Pi

# %&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&
# BASIC PHYSICAL FUNCTIONS
# %&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&

def bb_inu(temp_k=None, lam_cm=None):
    """
    Planck function. Returns f_nu given T and lambda.
    """
    nu = c / lam_cm
    fnu = 2.0*h*nu**3/c**2 /(_numpy.exp(h*nu/(kb*temp_k))-1.0)
    return fnu, "erg/s/cm^2/sr/Hz"

def bb_ilam(temp_k=None, lam_cm=None):
    """
    Planck function. Returns f_lambda given T and lambda.
    """
    flam = 2.0*h*c**2/lam_cm**5 / (_numpy.exp(h*c/(lam_cm*kb*temp_k))-1.0)
    return (flam, "erg/s/cm^2/sr/cm")

def peak_inu(temp_k=None):
    """
    Wavelength for peak of F_nu given T.
    """
    peak_lam = 2.821439372/kb/temp_k/h
    peak_nu = c / peak_lam
    return (peak_lam, "cm")

def peak_ilam(temp_k=None):
    """
    Wavelength for peak of F_lambda given T.
    """
    return h*c/kb/temp_k/4.965114231, "cm"

# %&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&
# UNIT CONVERSIONS
# %&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&
'''



"""
SIRIUS functions
"""
#__init__.py
from .calc_a_noise import calc_a_noise_chunk
from .calc_vis import calc_vis_chunk
from .calc_uvw import calc_uvw_chunk
from .calc_beam import calc_zpc_beam, evaluate_beam_models
from .simulation import simulation




#  CASA Next Generation Infrastructure
#  Copyright (C) 2021 AUI, Inc. Washington DC, USA
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
import numpy as np

def _check_parms(parm_dict, string_key, acceptable_data_types, acceptable_data = None, acceptable_range = None, list_acceptable_data_types=None, list_len=None, default=None):
    """

    Parameters
    ----------
    parm_dict: dict
        The dictionary in which the a parameter will be checked
    string_key :
    acceptable_data_types : list
    acceptable_data : list
    acceptable_range : list (length of 2)
    list_acceptable_data_types : list
    list_len : int
        If list_len is -1 than the list can be any length.
    default :
    Returns
    -------
    parm_passed : bool
        
    """
    
    import numpy as np

    if string_key in parm_dict:
        if (list in acceptable_data_types) or (np.array in acceptable_data_types):
            if (len(parm_dict[string_key]) != list_len) and (list_len != -1):
                print('######### ERROR:Parameter ', string_key, 'must be a list of ', list_acceptable_data_types, ' and length', list_len , '. Wrong length.')
                return False
            else:
                list_len = len(parm_dict[string_key])
                for i in range(list_len):
                    type_check = False
                    for lt in list_acceptable_data_types:
                        if isinstance(parm_dict[string_key][i], lt):
                            type_check = True
                    if not(type_check):
                          print('######### ERROR:Parameter ', string_key, 'must be a list of ', list_acceptable_data_types, ' and length', list_len, '. Wrong type of',type(parm_dict[string_key][i]))
                          return False
                          
                    if acceptable_data is not None:
                        if not(parm_dict[string_key][i] in acceptable_data):
                            print('######### ERROR: Invalid', string_key,'. Can only be one of ',acceptable_data,'.')
                            return False
                            
                    if acceptable_range is not None:
                        if (parm_dict[string_key][i] < acceptable_range[0]) or (parm_dict[string_key][i] > acceptable_range[1]):
                            print('######### ERROR: Invalid', string_key,'. Must be within the range ',acceptable_range,'.')
                            return False
        elif (dict in acceptable_data_types):
            parms_passed = True

            if default is None:
                print('######### ERROR:Dictionary parameters must have a default. Please report bug.')
                return False
            #print('is a dict',default)
            for default_element in default:
                if default_element in parm_dict[string_key]:
                    #print('1.*******')
                    #print(parm_dict[string_key], default_element, [type(default[default_element])], default[default_element])
                    if not(_check_parms(parm_dict[string_key], default_element, [type(default[default_element])], default=default[default_element])): parms_passed = False
                    #print('2.*******')
                else:
                    #print('parm_dict',default_element,string_key)
                    parm_dict[string_key][default_element] = default[default_element]
                    print('Setting default', string_key, '[\'',default_element,'\']', ' to ', default[default_element])
        else:
            type_check = False
            for adt in acceptable_data_types:
                if isinstance(parm_dict[string_key], adt):
                    type_check = True
            if not(type_check):
                print('######### ERROR:Parameter ', string_key, 'must be of type ', acceptable_data_types)
                return False
                    
            if acceptable_data is not None:
                if not(parm_dict[string_key] in acceptable_data):
                    print('######### ERROR: Invalid', string_key,'. Can only be one of ',acceptable_data,'.')
                    return False
                    
            if acceptable_range is not None:
                if (parm_dict[string_key] < acceptable_range[0]) or (parm_dict[string_key] > acceptable_range[1]):
                    print('######### ERROR: Invalid', string_key,'. Must be within the range ',acceptable_range,'.')
                    return False
    else:
        if default is not None:
            #print(parm_dict, string_key,  default)
            parm_dict[string_key] =  default
            print ('Setting default', string_key, ' to ', parm_dict[string_key])
        else:
            print('######### ERROR:Parameter ', string_key, 'must be specified')
            return False
            
    return True

def _check_dataset(vis_dataset,data_variable_name):
    try:
        temp = vis_dataset[data_variable_name]
    except:
        print ('######### ERROR Data array ', data_variable_name,'can not be found in dataset.')
        return False
    return True
    
    

def _check_sel_parms(xds,sel_parms,new_or_modified_data_variables={},required_data_variables={},drop_description_out_keys=[],append_to_in_id=False):
    """
    
    Parameters
    ----------
    xds : xarray.core.dataset.Dataset
        Input vis.zarr multi dataset.
    sel_parms : dictionary
    Returns
    -------
    psf_dataset : xarray.core.dataset.Dataset
    """
    import copy
    
    #If an empty xds is supplied
    if not('data_groups' in xds.attrs):
        if 'data_group_out_id' in sel_parms:
            xds.attrs['data_groups'] = [{str(sel_parms['data_group_out_id']):{'id':str(sel_parms['data_group_out_id'])}}]
        else:
            xds.attrs['data_groups'] = [{'0':{'id':'0'}}]
        append_to_in_id = True
    ######################
    
    data_group_ids = np.array(list(xds.data_groups[0])).astype(int)
    
    if 'data_group_in_id' in sel_parms:
        sel_parms['data_group_in'] = {'id':str(sel_parms['data_group_in_id'])}
    
    if 'data_group_out_id' in sel_parms:
        sel_parms['data_group_out'] = {'id':str(sel_parms['data_group_out_id'])}
        
    for mdv in new_or_modified_data_variables:
        if mdv in sel_parms:
            if 'data_group_out' in sel_parms:
                sel_parms['data_group_out'] = {**sel_parms['data_group_out'],**{mdv:sel_parms[mdv]}}
            else:
                sel_parms['data_group_out'] = {mdv:sel_parms[mdv]}
                
    if 'data_group_in' in sel_parms:
        if 'id' in sel_parms['data_group_in']:
            assert (int(sel_parms['data_group_in']['id']) in data_group_ids), "######### ERROR: data_group_in id does not exist in " + sel_parms['xds']
            sel_parms['data_group_in'] = copy.deepcopy(xds.data_groups[0][str(sel_parms['data_group_in']['id'])])
            print("Setting data_group_in  to ",sel_parms['data_group_in'])
            data_group_in_default = sel_parms['data_group_in']
            
            if (append_to_in_id==True):
                new_data_group_id = sel_parms['data_group_in']['id']
            else:
                new_data_group_id = np.max(data_group_ids) + 1
        else:
            data_group_in_default = list(xds.data_groups[0])[0]
            data_group_in_default['id'] = -1 #Custom
    else:
        data_group_in_default = list(xds.data_groups[0].values())[0] #Choose id 1
        
        if (append_to_in_id==True):
            new_data_group_id = data_group_in_default['id']
        else:
            new_data_group_id = np.max(data_group_ids) + 1
            
    if 'data_group_out_id' in sel_parms:
        new_data_group_id = sel_parms['data_group_out_id']
        
    #If new coord is created new_or_modified_data_variables should contain all data variables that will be present in the output
    if 'data_group_out' in sel_parms:
        temp_sel_parms = copy.deepcopy(sel_parms)
        for sel in temp_sel_parms['data_group_out']:
            if not(sel in new_or_modified_data_variables) and (sel!='id'):
                sel_parms['data_group_out'].pop(sel,None)
                
    data_group_in_default = {**required_data_variables,**data_group_in_default}
    
    data_group_out_defaults = {**{'id':str(new_data_group_id)} , **new_or_modified_data_variables}
    #This is for type add_to_xds
    #print(data_group_in_default,data_group_out_defaults)
    data_group_out_defaults = {**data_group_in_default,**data_group_out_defaults} #{**x, **y} merges dicts with y taking precedence for repeated entries.
    
    sel_defaults = {'data_group_in':data_group_in_default,'data_group_out':data_group_out_defaults}
    assert(_check_sub_sel_parms(sel_parms,sel_defaults)), "######### ERROR: sel_parms checking failed"
    
    sel_check = {'data_group_in':sel_parms['data_group_in']}
    assert(_check_existence_sel_parms(xds,sel_check)), "######### ERROR: sel_parms checking failed"
    
    #data_group_id = sel_parms['data_group_out']['id']
    
    
    #print('%%%%%%%%%%')
    #print(sel_parms['data_group_in'], data_group_in_default)
    
    
    for mdv in new_or_modified_data_variables:
        for d_id in xds.data_groups[0]:
            #print(xds.data_groups[0][d_id][modified_data_variable])
            
            if mdv in xds.data_groups[0][d_id]:
                assert (sel_parms['data_group_out']['id'] == d_id) or (xds.data_groups[0][d_id][mdv] != sel_parms['data_group_out'][mdv]), "Data variables, that are modified by the function, can not be replaced if they are referenced in another data_group"
            #if (sel_parms['data_group_out']['id'] != d_id) and (xds.data_groups[0][d_id][modified_data_variable] == sel_parms['data_group_out'][modified_data_variable]):
            #    print(sel_parms['data_group_out']['id'], d_id)
        
    #for drop_description_out_keys
    #print('AAAAA',sel_parms)
    return True
    
    
    
def _check_sub_sel_parms(sel_parms,select_defaults):
    parms_passed = True
    for sel in select_defaults:
        if not(_check_parms(sel_parms, sel, [type(select_defaults[sel])], default=select_defaults[sel])): parms_passed = False
    return parms_passed

'''
def _check_sel_parms(sel_parms,select_defaults):
    parms_passed = True
    for sel_def in select_defaults:
        if isinstance(select_defaults[sel_def], dict):
            if sel_def in sel_parms:
                for sub_sel_def in select_defaults[sel_def]:
                        #print(sub_sel_def,select_defaults[sel_def])
                        #print(sel_parms[sel_def], sub_sel_def, select_defaults[sel_def][sub_sel_def])
                        if not(_check_parms(sel_parms[sel_def], sub_sel_def, [str], default=select_defaults[sel_def][sub_sel_def])): parms_passed = False
            else:
                sel_parms[sel_def] = select_defaults[sel_def]
                print ('Setting default', string_key, ' to ', parm_dict[string_key])
        else:
            if not(_check_parms(sel_parms, sel_def, [str], default=select_defaults[sel_def])): parms_passed = False
    return parms_passed
'''

def _check_existence_sel_parms(dataset, sel_parms):
    parms_passed = True
    for sel in sel_parms:
        if isinstance(sel_parms[sel], dict):
            if sel != 'properties':
                _check_existence_sel_parms(dataset, sel_parms[sel])
        else:
            if (sel != 'id') and (sel != 'properties'):
                if not(_check_dataset(dataset,sel_parms[sel])): parms_passed = False
    return parms_passed



#   Copyright 2019 AUI, Inc. Washington DC, USA
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import numpy as np
from sirius_data._constants import c
import scipy.constants
from scipy.special import j1, jn
from numba import jit
import numba_scipy.special

# The twiddle factor is needed to improve agreement between CASA PBMATH airy disk and SiRIUS. 
# This arrises because CASA makes use of truncated constants. The twiddle factor 0.9998277835716939 is very close to 1. 
# Even with the twiddle there will not be machine precission agreement, because CASA calculates a 1D PB at 1 GHz for 10000 points and then frequency scales and interpolates to desired value.
casa_twiddle = (180*7.016*c)/((np.pi**2)*(10**9)*1.566*24.5) # 0.9998277835716939
r_vla_max = 0.8564*np.pi/180 #in radians

#@jit(nopython=True,cache=True,nogil=True)
@jit(nopython=True,nogil=True)
def _casa_airy_pb(lmn,freq_chan,dish_diameter, blockage_diameter, ipower, r_max=r_vla_max, n_sample=10000):
    #dish_diameter = pb_parms['dish_diameter']
    #blockage_diameter = pb_parms['blockage_diameter']
    #ipower = pb_parms['ipower']
    
    #print(lmn,freq_chan,dish_diameter, blockage_diameter, ipower, r_max, n_sample)
    
    k = (2*np.pi*freq_chan)/c
    aperture = dish_diameter/2

    if n_sample is not None:
        r = np.arcsin(np.sqrt(lmn[0]**2 + lmn[1]**2)) #CASA in PBMATH does a small angle approximation.
        r_inc = ((r_max)/(n_sample-1))
        r = (int(np.floor(r/r_inc + 0.5))*r_inc)*aperture*k
        r = r*casa_twiddle
    else:
        r = np.arcsin(np.sqrt(lmn[0]**2 + lmn[1]**2)*k*aperture)
        
    if (r != 0):
        if blockage_diameter==0.0:
            return (2.0*j1(r)/r)**ipower
        else:
            area_ratio = (dish_diameter/blockage_diameter)**2
            length_ratio = (dish_diameter/blockage_diameter)
            return ((area_ratio * 2.0 * j1(r)/r   - 2.0 * j1(r * length_ratio)/(r * length_ratio) )/(area_ratio - 1.0))**ipower
    else:
        return 1

#@jit(nopython=True,cache=True,nogil=True)
@jit(nopython=True,nogil=True)
def _airy_pb(lmn,freq_chan,dish_diameter, blockage_diameter, ipower):
    #print('lmn is',lmn)
    
    if (lmn[0] != 0) or (lmn[1] != 0):
        #dish_diameter = pb_parms['dish_diameter']
        #blockage_diameter = pb_parms['blockage_diameter']
        #ipower = pb_parms['ipower']
        
        k = (2*np.pi*freq_chan)/c
        
        aperture = dish_diameter/2
        r = np.sqrt(lmn[0]**2 + lmn[1]**2)*k*aperture
        
        if blockage_diameter==0.0:
            return (2.0*j1(r)/r)**ipower
        else:
            e = blockage_diameter/dish_diameter
            return (( 2.0 * j1(r)/r   - 2.0 * e * j1(r * e)/r )/(1.0 - e**2))**ipower
            #Changed r_grid to r ^^^
    else:
        return 1
    
    
    
#Non-jitted version:
def _casa_airy_pb_njit(lmn,freq_chan,pb_parms):
    #print('lmn is',lmn)
    
    if (lmn[0] != 0) or (lmn[1] != 0):
        dish_diameter = pb_parms['dish_diameter']
        blockage_diameter = pb_parms['blockage_diameter']
        ipower = pb_parms['ipower']
        
        k = (2*np.pi*freq_chan)/c
        
        aperture = dish_diameter/2
        r = np.sqrt(lmn[0]**2 + lmn[1]**2)*k*aperture
        
        if blockage_diameter==0.0:
            return (2.0*jn(1,r)/r)**ipower
        else:
            area_ratio = (dish_diameter/blockage_diameter)**2
            length_ratio = (dish_diameter/blockage_diameter)
            return ((area_ratio * 2.0 * jn(1,r)/r   - 2.0 * jn(1, r * length_ratio)/(r * length_ratio) )/(area_ratio - 1.0))**ipower
    else:
        return 1


def _airy_pb_njit(lmn,freq_chan,pb_parms):
    #print('lmn is',lmn)
    
    if (lmn[0] != 0) or (lmn[1] != 0):
        dish_diameter = pb_parms['dish_diameter']
        blockage_diameter = pb_parms['blockage_diameter']
        ipower = pb_parms['ipower']
        
        k = (2*np.pi*freq_chan)/c
        
        aperture = dish_diameter/2
        r = np.sqrt(lmn[0]**2 + lmn[1]**2)*k*aperture
        
        if blockage_diameter==0.0:
            return (2.0*jn(1,r)/r)**ipower
        else:
            e = blockage_diameter/dish_diameter
            return (( 2.0 * jn(1,r)/r   - 2.0 * e * jn(1, r * e)/r )/(1.0 - e**2))**ipower
    else:
        return 1



#   Copyright 2019 AUI, Inc. Washington DC, USA
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import numpy as np
import time
from numba import jit
import numba
from ._sirius_utils._coord_transforms import _calc_rotation_mats, _directional_cosine,  _sin_project
from ._sirius_utils._beam_utils import _calc_pb_scale, _beam_models_to_tuple, _pol_code_to_index
from ._sirius_utils._array_utils import _is_subset
from sirius_data._constants import map_mueler_to_pol, c
from ._parm_utils._check_parms import _check_parms
#def calc_vis(uvw,vis_data_shape,point_source_flux,point_source_ra_dec,pointing_ra_dec,phase_center_ra_dec,antenna1,antenna2,freq_chan,beam_model_map,beam_models, parallactic_angle, pol, mueller_selection):
#    return 0

def calc_vis_chunk(uvw,vis_data_shape, point_source_flux,point_source_ra_dec,pointing_ra_dec,phase_center_ra_dec,antenna1,antenna2,chan_chunk,beam_model_map,beam_models, parallactic_angle, pol, mueller_selection, check_parms=True):
    """
    Simulate interferometric visibilities.
    
    Parameters
    ----------
    uvw : float np.array, [n_time,n_baseline,3]   
        Spatial frequency coordinates.
    vis_data_shape : float np.array, [4]
        Dimensions of visibility data [n_time, n_baseline, n_chan, n_pol].
    point_source_flux: float np.array, [n_point_sources,n_time, n_chan, n_pol], (singleton: n_time, n_chan), Janskys
        The flux of the point sources.
    point_source_ra_dec: float np.array, [n_time, n_point_sources, 2], (singleton: n_time), radians
        The position of the point sources.
    pointing_ra_dec: float np.array, [n_time, n_ant, 2], (singleton: n_time, n_ant), radians
        Pointings of antennas, if they are different from the phase center. Set to None if no pointing offsets are required.
    phase_center_ra_dec: float np.array, [n_time, 2], (singleton: n_time), radians
        Phase center of array.
    antenna1: np.array of int, [n_baseline]
        The indices of the first antenna in the baseline pairs. The _calc_baseline_indx_pair function in sirius._sirius_utils._array_utils can be used to calculate these values.
    antenna2: np.array of int, [n_baseline]
        The indices of the second antenna in the baseline pairs. The _calc_baseline_indx_pair function in sirius._sirius_utils._array_utils can be used to calculate these values.
    chan_chunk: float np.array, [n_chan], Hz
        Channel frequencies.
    beam_model_map: int np.array, [n_ant]
        Each element in beam_model_map is an index into beam_models.
    beam_models: list
        List of beam models to use. Beam models can be any combination of function parameter dictionaries or image xr.Datasets. Any Zernike polynomial coefficient xr.Datasets models must be converted to images using sirius.calc_beam.evaluate_beam_models.
    parallactic_angle: float np.array, [n_time], radians
        Parallactic angle over time at the array center.
    pol: int np.array 
        Must be a subset of ['RR','RL','LR','LL'] => [5,6,7,8] or ['XX','XY','YX','YY'] => [9,10,11,12].
    mueller_selection: int np.array
        The elements in the 4x4 beam Mueller matrix to use. The elements are numbered row wise.
        For example [ 0, 5, 10, 15] are the diagonal elements.
    check_parms: bool
        Check input parameters and asign defaults.
        
    Returns
    -------
    vis : complex np.array, [n_time,n_baseline,n_chan,n_pol]   
        Visibility data.
    """
    
    if check_parms:
        n_time, n_baseline, n_chan, n_pol = vis_data_shape
        n_ant = tel_xds.dims['ant_name']
        
        pol = np.array(pol)
        assert(_is_subset(pol_codes_RL,pol) or _is_subset(pol_codes_XY,pol)), 'Pol selection invalid, must either be subset of [5,6,7,8] or [9,10,11,12] but is '
        if not(_check_parms(beam_parms, 'mueller_selection', [list,np.array], list_acceptable_data_types=[np.int64], default = np.array([ 0, 5, 10, 15]),list_len=-1)): parms_passed = False
        
        #Check dimensions.
        assert(point_source_flux.shape[0] == point_source_ra_dec.shape[1]), 'n_point_sources dimension of point_source_flux[' + str(point_source_flux.shape[0]) +'] and point_source_ra_dec['+str(point_source_ra_dec.shape[1])+'] do not match.'
        assert(point_source_flux.shape[1] == 1) or (point_source_flux.shape[1] == n_time), 'n_time dimension in point_source_flux[' + str(point_source_flux.shape[1]) + '] must be either 1 or ' + str(n_time) + ' (see time_xda parameter).'
        assert(point_source_flux.shape[2] == 1) or (point_source_flux.shape[2] == n_chan), 'n_chan dimension in point_source_flux[' + str(point_source_flux.shape[2]) + '] must be either 1 or ' + str(n_chan) + ' (see chan_xda parameter).'
        assert(point_source_flux.shape[3] == 4), 'n_pol dimension in point_source_flux[' + str(point_source_flux.shape[3]) + '] must be 4.'
        
        assert(point_source_ra_dec.shape[0] == 1) or (point_source_ra_dec.shape[0] == n_time), 'n_time dimension in point_source_ra_dec[' + str(point_source_ra_dec.shape[0]) + '] must be either 1 or ' + str(n_time) + ' (see time_xda parameter).'
        assert(point_source_ra_dec.shape[2] == 2), 'ra,dec dimension in point_source_ra_dec[' + str(point_source_ra_dec.shape[2]) + '] must be 2.'
        
        if pointing_ra_dec is not None:
            assert(pointing_ra_dec.shape[0] == 1) or (pointing_ra_dec.shape[0] == n_time), 'n_time dimension in pointing_ra_dec[' + str(pointing_ra_dec.shape[0]) + '] must be either 1 or ' + str(n_time) + ' (see time_xda parameter).'
            assert(pointing_ra_dec.shape[1] == 1) or (pointing_ra_dec.shape[1] == n_ant), 'n_ant dimension in pointing_ra_dec[' + str(pointing_ra_dec.shape[1]) + '] must be either 1 or ' + str(n_ant) + ' (see tel_xds.dims[\'ant_name\']).'
            assert(pointing_ra_dec.shape[2] == 2), 'ra,dec dimension in pointing_ra_dec[' + str(pointing_ra_dec.shape[2]) + '] must be 2.'
            
            
        assert(phase_center_ra_dec.shape[0] == 1) or (phase_center_ra_dec.shape[0] == n_time), 'n_time dimension in phase_center_ra_dec[' + str(phase_center_ra_dec.shape[0]) + '] must be either 1 or ' + str(n_time) + ' (see time_xda parameter).'
        assert(phase_center_ra_dec.shape[1] == 2), 'ra,dec dimension in phase_center_ra_dec[' + str(phase_center_ra_dec.shape[1]) + '] must be 2.'
            
        assert(phase_center_names.shape[0] == 1) or (phase_center_names.shape[0] == n_time), 'n_time dimension in phase_center_ra_dec[' + str(phase_center_names.shape[0]) + '] must be either 1 or ' + str(n_time) + ' (see time_xda parameter).'
    
    # The _beam_models_to_tuple function is here to appease Numba the terrible. It unpacks the beam models from dictionaries and xr.Datasets to fixed tuples.
    beam_models_type0, beam_models_type1, beam_types, new_beam_model_map = _beam_models_to_tuple(beam_models,beam_model_map) 
    
    #print(beam_models_type0, beam_models_type1, beam_types, new_beam_model_map)
    
    do_pointing = False
    if pointing_ra_dec is not None:
        do_pointing = True
    else:
        pointing_ra_dec = np.zeros((2,2,2))
        
    vis_data = np.zeros(vis_data_shape,dtype=np.complex128)
    calc_vis_jit(vis_data, uvw,tuple(vis_data_shape),point_source_flux.astype(np.complex128),point_source_ra_dec,pointing_ra_dec,phase_center_ra_dec,antenna1,antenna2,chan_chunk,beam_models_type0, beam_models_type1, beam_types, new_beam_model_map, parallactic_angle, pol, mueller_selection, do_pointing)
    
    return vis_data
    
    
#@jit(nopython=True,cache=True,nogil=True)
@jit(nopython=True,nogil=True) #Jit compile function because it has large nested for loop (can't be easily vectorized).
def calc_vis_jit(vis_data,uvw,vis_data_shape,point_source_flux,point_source_ra_dec,pointing_ra_dec,phase_center_ra_dec,antenna1,antenna2,freq_chan,beam_models_type0, beam_models_type1, beam_types, beam_model_map, parallactic_angle, pol, mueller_selection, do_pointing):

    n_time, n_baseline, n_chan, n_pol = vis_data_shape
    n_ant = len(beam_model_map)

    n_point_source = point_source_ra_dec.shape[1]
    
    #Converts pol codes to indices.
    pol = _pol_code_to_index(pol)
    
    #Find all singleton dimensions, so that indexing is done correctly.
    n_time, n_baseline, n_chan, n_pol = vis_data_shape
    n_ant = len(beam_model_map)
    f_pc_time = n_time if phase_center_ra_dec.shape[0] == 1 else 1
    f_ps_time = n_time if point_source_ra_dec.shape[0] == 1 else 1
    f_sf_time = n_time if point_source_flux.shape[1] == 1 else 1
    f_sf_chan = n_chan if point_source_flux.shape[2] == 1 else 1
    
  
    if do_pointing:
        f_pt_time = n_time if phase_center_ra_dec.shape[0] == 1 else 1
        f_pt_ant =  n_ant if point_source_ra_dec.shape[1] == 1 else 1
    else:
        f_pt_time = n_time
        f_pt_ant = n_ant
    
    #prev_ra_dec_in =  np.zeros((4,))
    #prev_ra_dec_out = np.zeros((4,))
    prev_ra_dec_in =  np.zeros((4,),dtype=numba.float64)
    prev_ra_dec_out = np.zeros((4,),dtype=numba.float64)
    
    f_pc_time = n_time if phase_center_ra_dec.shape[0] == 1 else 1
    
    # Loop over dimensions n_time, n_baseline, n_point_source, n_chan, n_pol
    for i_time in range(n_time):
        #print("Completed time step ", i_time,"of",n_time)
        pa = parallactic_angle[i_time]
        ra_dec_in = phase_center_ra_dec[i_time//f_pc_time, :]
        
        for i_baseline in range(n_baseline):
            i_ant_1 = antenna1[i_baseline]
            i_ant_2 = antenna2[i_baseline]
            if do_pointing:
                ra_dec_in_1 = pointing_ra_dec[i_time//f_pt_time,i_ant_1//f_pt_ant,:]
                ra_dec_in_2 = pointing_ra_dec[i_time//f_pt_time,i_ant_2//f_pt_ant,:]
                 
            for i_point_source in range(n_point_source):
                ra_dec_out = point_source_ra_dec[i_time//f_ps_time,i_point_source,:]
                if not(np.array_equal(prev_ra_dec_in, ra_dec_in) and np.array_equal(prev_ra_dec_out, ra_dec_out)):
                    uvw_rotmat, lmn_rot = _calc_rotation_mats(ra_dec_in, ra_dec_out)
                    lm_sin = _sin_project(ra_dec_in,ra_dec_out)
                    sep = np.sqrt(np.sum(lm_sin**2))
                    
                if do_pointing:
                    if not(np.array_equal(prev_ra_dec_in, ra_dec_in_1) and np.array_equal(prev_ra_dec_out, ra_dec_out)):
                        lm_sin_1 = _sin_project(ra_dec_in_1, ra_dec_out) 
                        sep_1 = np.sqrt(np.sum(lm_sin_1**2))
                    if not(np.array_equal(prev_ra_dec_in, ra_dec_in_2) and np.array_equal(prev_ra_dec_out, ra_dec_out)):
                        lm_sin_2 = _sin_project(ra_dec_in_2, ra_dec_out)
                        sep_2 = np.sqrt(np.sum(lm_sin_2**2))
                

                phase = 2*np.pi*lmn_rot@(uvw[i_time,i_baseline,:]@uvw_rotmat)
                prev_ra_dec_in = ra_dec_in
                prev_ra_dec_out = ra_dec_out

                for i_chan in range(n_chan):
                    #s1 = time.time()
                    flux = point_source_flux[i_point_source,i_time//f_sf_time, i_chan//f_sf_chan, :]
                    #print("s1",time.time()-s1)
                    
                    ################### Apply primary beams #####################
                    bm1_indx = beam_model_map[i_ant_1]
                    bm2_indx = beam_model_map[i_ant_2]
                    bm1_type = beam_types[i_ant_1]
                    bm2_type = beam_types[i_ant_2]

                    #s2 = time.time()
                    if do_pointing:
                        flux_scaled, outside_beam = _calc_pb_scale(flux,sep_1,sep_2,bm1_indx,bm2_indx,bm1_type,bm2_type,lm_sin_1,lm_sin_2,beam_models_type0,beam_models_type1,pa,freq_chan[i_chan],mueller_selection,do_pointing)
                    else:
                        flux_scaled, outside_beam = _calc_pb_scale(flux,sep,sep,bm1_indx,bm2_indx,bm1_type,bm2_type,lm_sin,lm_sin,beam_models_type0,beam_models_type1,pa,freq_chan[i_chan],mueller_selection,do_pointing)
                    #print("s2",time.time()-s2)
                    
                    if not outside_beam:
                        #s3 = time.time()
                        phase_scaled = 1j*phase*freq_chan[i_chan]/c
                        #print("s3",time.time()-s3)
                        
                        #s4 = time.time()
                        for i_pol in range(n_pol):
                            vis_data[i_time,i_baseline,i_chan,i_pol] = vis_data[i_time,i_baseline,i_chan,i_pol] + flux_scaled[pol[i_pol]]*np.exp(phase_scaled)/(1-lmn_rot[2])
                        #print("s4",time.time()-s4)





import numpy as np
from sirius._sirius_utils._math_utils import _powl2, _powl

def test_powl2():
    assert np.allclose(_powl2(np.array([[5.]]), 5), np.array([[5.**5]])) == True
    
def test_powl2_array():
    assert np.allclose(_powl2(np.array([[5., 6., 7.]]), 5), np.array([[5., 6., 7.]])**5) == True
    
def test_powl2_arrays():
    assert np.allclose(_powl2(np.array([[5., 6., 7.], [5., 6., 7.]]), 5), np.array([[5., 6., 7.], [5., 6., 7.]])**5) == True
    
def test_powl():
    assert np.allclose(_powl(5., 5), 5**5) == True


#  CASA Next Generation Infrastructure
#  Copyright (C) 2021 AUI, Inc. Washington DC, USA
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from numba import jit, types
import numba
import numpy as np

@jit(numba.float64[::1](types.Array(types.float64, 2, 'A', readonly=True), numba.float64[::1], numba.float64[::1]), nopython=True,cache=True,nogil=True)
def _bilinear_interpolate(im, x, y):
    """Interpolates image values. 
    Inputs 
    -------------- 
    im: 2-d numpy array (complex?)
    x: 1-d numpy array of fractional indices (float)
    y: 1-d numpy array of fractional indices (float)
    Notes: x and y must be same length. Negative indices not allowed (automatically set to 0).
    -------------- 
    Outputs: 
    -------------- 
    1-d numpy array of interpolated values (float)"""
    
    #x0 rounds down, x1 rounds up. Likewise for y
    x0 = np.floor(x).astype(numba.intc)
    x1 = x0 + 1
    y0 = np.floor(y).astype(numba.intc)
    y1 = y0 + 1
    
    #Safety: makes sure no indices out of bounds
    x0 = np.minimum(im.shape[1]-1, np.maximum(x0, 0))
    x1 = np.minimum(im.shape[1]-1, np.maximum(x1, 0))
    y0 = np.minimum(im.shape[0]-1, np.maximum(y0, 0))
    y1 = np.minimum(im.shape[0]-1, np.maximum(y1, 0))
    
    #Four values around value to be interpolated
    Ia1 = im[y0]
    Ia = Ia1.flatten()[x0]
    Ib1 = im[y1]
    Ib = Ib1.flatten()[x0]
    Ic1 = im[y0]
    Ic = Ic1.flatten()[x1]
    Id1 = im[y1]
    Id = Id1.flatten()[x1]
    
    #See https://en.wikipedia.org/wiki/Bilinear_interpolation
    wa = (x1-x) * (y1-y)
    wb = (x1-x) * (y-y0)
    wc = (x-x0) * (y1-y)
    wd = (x-x0) * (y-y0)

    return wa*Ia + wb*Ib + wc*Ic + wd*Id


@jit(types.Array(types.complex128, 2, 'C')(types.Array(types.complex128, 3, 'C'), types.Array(types.float64, 1, 'C'), types.Array(types.float64, 1, 'C'), types.float64, types.float64), nopython=True,cache=True,nogil=True)
def _interp_array(im_array, l, m, delta_l, delta_m):
    """Interpolates image values.
    Inputs 
    -------------- 
    im_array: 3-d numpy array of shape (pol, image, image)
    l: 1-d numpy array of fractional indices (float)
    m: 1-d numpy array of fractional indices (float)
    delta_l: pixel size for l coordinates (float)
    delta_m: pixel size for m coordinates (float)
    Notes: l and m must be same length.
    -------------- 
    Outputs: 
    -------------- 
    2-d numpy array of interpolated values (float) with shape (pol, len(l))"""
    #Length of image along l 
    n_l = len(im_array[0, :, 0]) #Change to shape?
    #Length of image along m
    n_m = len(im_array[0, 0, :]) 
    #Fractional pixel along l
    x_frac = (l/delta_l) + n_l//2 
    #Fractional pixel along m
    y_frac = (m/delta_m) + n_m//2 
    #Numba-style array creation. Shape is (polarization, coordinates)
    results = np.zeros((len(im_array), len(l)), dtype = numba.complex128) 
    for i in range(len(im_array)):
        #Complex interpolation
        results[i] = _bilinear_interpolate(im_array[i].real, x_frac, y_frac) +    1j*_bilinear_interpolate(im_array[i].imag, x_frac, y_frac)
    return results.astype(numba.complex128)


"""
@jit(nopython=True,cache=True,nogil=True)
def interp_ndim(ndim_array, x, y, dims = (0, 1)):
    Interpolates coordinates of an image defined by specified dimensions of an n-d array.
    Inputs:
    ndim_array: n-dimensional array
    dims: tuple of size 2 containing dimensions which comprise the image
    Outputs:
    Interpolated values
    #Gets the shape of the images which are to be interpolated
    shape = ndim_array.shape
    shape_image = np.zeros(2, dtype = int)
    for i in range(2):
        shape_image[i] = shape[dims[i]]
    
    #Gets the shape of the container of the images
    shape_image_array = np.delete(ndim_array.shape, dims)
    shape_image_tuple = (shape_image_array[0], )
    for i in range(1, len(shape_image_array)):
        shape_image_tuple = shape_image_tuple + (shape_image_array[i], )
    #print(shape_image_tuple)
    
    #Creates flattened container with images inside
    length_flat = 1
    for i in range(len(shape_image_array)):
        length_flat = length_flat*shape_image_array[i]
    flat_shape = np.zeros(1, dtype = int)
    flat_shape = (length_flat, shape_image[0], shape_image[1])
    f_image_array = ndim_array.reshape(flat_shape)
    
    #Creates container with results of interpolation as the innermost dimension
    results = np.zeros((len(f_image_array), len(x)), dtype = float)
    for i in range(len(f_image_array)):
        results[i] = _bilinear_interpolate(f_image_array[i], x, y)
    shape_results = shape_image_tuple + (len(x), )
    results = results.reshape(shape_results)
    return results
            
"""

@jit(numba.float64[:, :](numba.float64[:,:], numba.int64), nopython=True, cache=True, nogil=True)
def _powl2(base_arr, in_exp):
    #print(base,exp)
    #base can be any real and exp must be positive integer
    # base**exp
    """
    Algorithm taken from https://stackoverflow.com/questions/2198138/calculating-powers-e-g-211-quickly
    https://en.wikipedia.org/wiki/Exponentiation_by_squaring
    """
    #exp_int = np.zeros(base_arr.shape)
    exp_int = np.zeros(base_arr.shape,numba.f8)
    
    for i in range(base_arr.shape[0]):
        for j in range(base_arr.shape[1]):
            base = base_arr[i,j]
            
            exp = in_exp
            r = 0.0
            if exp == 0:
                r = 1.0
            else:
                if exp < 0 :
                    base = 1.0 / base
                    exp = -exp
                 
                y = 1.0;
                while exp > 1:
                    if (exp % 2) == 0:  #exp is even
                    #if (exp & 1) == 0:
                        base = base * base
                        #exp = exp / 2
                    else:
                        y = base * y
                        base = base * base
                    #exp = (exp – 1) / 2
                    exp = exp//2
                r = base * y
            exp_int[i,j] = r
    
    return exp_int

@jit(types.Array(types.float64, 2, 'C')(types.Array(types.float64, 2, 'C'), numba.int64), nopython=True, cache=True, nogil=True)
def _powl(base, exp):
    #print(base,exp)
    #base can be any real and exp must be positive integer
    """
    Algorithm taken from https://stackoverflow.com/questions/2198138/calculating-powers-e-g-211-quickly
    https://en.wikipedia.org/wiki/Exponentiation_by_squaring
    """
    if exp == 0:
        return np.ones(base.shape)
    elif exp == 1:
        return base
    elif (exp & 1) != 0: #is even
        return base * _powl(base * base, exp // 2)
    else:
        return _powl(base * base, exp // 2)


@jit(numba.float64(numba.float64[:,:], numba.float64[:,:]), nopython=True,cache=True,nogil=True)
def mat_dis(A,B):
    return(np.sum(np.abs(A-B)))


##################Currently Not used functions##################
#@jit(nopython=True,cache=True)
def _outer_product(B1,B2,norm,conj):
    '''
    Input
    B1 2 x 2 x m x n array
    B2 2 x 2 x m x n array
    Output
    M 4 x 4 x m x n
    '''
    
    #assert B1.shape==B2.shape
    
    s = B1.shape
    
    M = np.zeros((4,4,s[2],s[3]),dtype=np.complex128)
    
    indx_b1 = np.array([[[0,0],[0,0],[0,1],[0,1]],[[0,0],[0,0],[0,1],[0,1]],[[1,0],[1,0],[1,1],[1,1]],[[1,0],[1,0],[1,1],[1,1]]])
    indx_b2 = np.array([[[0,0],[0,1],[0,0],[0,1]],[[1,0],[1,1],[1,0],[1,1]],[[0,0],[0,1],[0,0],[0,1]],[[1,0],[1,1],[1,0],[1,1]]])
    #print(indx_b1.shape)
    
    
    for i in range(4):
        for j in range(4):
            #print(indx_b1[i,j,:], ',*,', indx_b2[i,j,:])
            if conj:
                M[i,j,:,:] = B1[indx_b1[i,j,0],indx_b1[i,j,1],:,:] * B2[indx_b2[i,j,0],indx_b2[i,j,1],:,:].conj().T
            else:
                M[i,j,:,:] = B1[indx_b1[i,j,0],indx_b1[i,j,1],:,:] * B2[indx_b2[i,j,0],indx_b2[i,j,1],:,:]
                
            if norm:
                M[i,j,:,:] = M[i,j,:,:]/np.max(np.abs(M[i,j,:,:]))
    
    #print(M.shape)
    return(M)


def _outer_product_conv(B1,B2):
    
#    Input
#    B1 2 x 2 x m x n array
#    B2 2 x 2 x m x n array
#    Output
#    M 4 x 4 x m x n
    
    #assert B1.shape==B2.shape
    
    s = B1.shape
    
    M = np.zeros((4,4,s[2],s[3]),dtype=np.complex128)
    
    indx_b1 = np.array([[[0,0],[0,0],[0,1],[0,1]],[[0,0],[0,0],[0,1],[0,1]],[[1,0],[1,0],[1,1],[1,1]],[[1,0],[1,0],[1,1],[1,1]]])
    indx_b2 = np.array([[[0,0],[0,1],[0,0],[0,1]],[[1,0],[1,1],[1,0],[1,1]],[[0,0],[0,1],[0,0],[0,1]],[[1,0],[1,1],[1,0],[1,1]]])
    
    for i in range(4):
        for j in range(4):
            M[i,j,:,:] = signal.fftconvolve(B1[indx_b1[i,j,0],indx_b1[i,j,1],:,:], B2[indx_b2[i,j,0],indx_b2[i,j,1],:,:],mode='same')
    
    print(M.shape)
    return(M)

    
def _make_flat(B):
    '''
    B 2x2xmxn
    B_flat 2mx2n
    '''
    s = B.shape
    B_flat = np.zeros((s[2]*s[0],s[3]*s[1]),dtype=complex)
    
    
    for i in range(s[0]):
        for j in range(s[1]):
            i_start = i*s[2]
            i_end = (i+1)*s[3]
            j_start = j*s[2]
            j_end = (j+1)*s[3]
            B_flat[i_start:i_end,j_start:j_end] = B[i,j,:,:]
            #print(B[i,j,1024,1024],np.abs(B[i,j,1024,1024]))
    return B_flat
    
    
def _make_flat_casa(B):
    '''
    B mxnx16
    B_flat 4mx4n
    '''
    s = B.shape
    B_flat = np.zeros((s[0]*4,s[1]*4),dtype=complex)
    
    #indx = np.array([[0,0],[1,0],[2,0],[3,0],[0,1],[1,1],[2,1],[3,1],[0,2],[1,2],[2,2],[3,2],[0,3],[1,3],[2,3],[3,3]])
    indx = np.array([[0,0],[0,1],[0,2],[0,3],[1,0],[1,1],[1,2],[1,3],[2,0],[2,1],[2,2],[2,3],[3,0],[3,1],[3,2],[3,3]]) #saved as rows
    
    for c,i in enumerate(indx):
        #print(c,i)
        i_start = i[0]*s[0]
        i_end = (i[0]+1)*s[0]
        j_start = i[1]*s[1]
        j_end = (i[1]+1)*s[1]
        B_flat[i_start:i_end,j_start:j_end] = B[:,:,c].T
        #print(B[1024,1024,c],np.abs(B[1024,1024,c]))
    return B_flat

